unit MoleculeRenderer;

{$ifdef FPC}
  {$mode DELPHI}
{$endif}

{ 2D molecule renderer that parses molecule rendering info V2 produced by
  Proteax and generates corresponding graphical output on a standard TCanvas.

(C) 2009, 2010 Jan Holst Jensen, jan@biochemfusion.com.

This pascal source code unit (MoleculeRenderer.pas) is released under a BSD-style
license:

* Copyright (C) 2009-2010, Biochemfusion (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

}

interface

uses
  {$ifndef FPC} Windows, {$endif} Classes, SysUtils, Graphics;

type
  EMoleculeRenderer = class(Exception);

  TMoleculeRenderer = class
  protected
    XMax: Integer;
    YMax: Integer;
    //
    CoordScaling: Integer;
    BoxWidth: Integer;
    BoxHeight: Integer;
    //
    fOneLetterResidues: Boolean;
    // Template string for measuring text width of one- or three-letter residue symbol circles.
    fResidueSymbolTemplate: String;
    fResidueCircleRadius: Integer;
    // Size difference between normal atoms and residue atoms. Depends on whether residue atoms
    // are shown as one- or three-letter symbols.
    fHasCalculatedResidueFontSize: Boolean;
    fResidueFontSizeDelta: Integer;
    procedure Check(Condition: Boolean; const ErrMsg: String);
    procedure ProcessRenderInfo(const RenderInfo: String; Rect: TRect; const ZoomFactor: Double; ACanvas: TCanvas; WhatToDo: Integer);
  public
    function  BoundingBox(const RenderInfo: String; const ZoomFactor: Double; ACanvas: TCanvas): TRect;
    procedure Render(const RenderInfo: String; Rect: TRect; const ZoomFactor: Double; ACanvas: TCanvas);

    property OneLetterResidues: Boolean read fOneLetterResidues write fOneLetterResidues;
  end;

implementation

uses
  Math, MoleculeRendererConsts;

var
  _ThreeLetterCodes: TStringList;
  _OneLetterCodes: String;

const
  DO_RENDER = 0;
  DO_CALCBB = 1;

  NUM_CHAIN_BACKGND_COLORS = 2;
  CHAIN_BACKGND_COLORS: array [0..NUM_CHAIN_BACKGND_COLORS - 1] of TColor =
    (clMoneyGreen, clSkyBlue);

{ TMoleculeRenderer }

procedure TMoleculeRenderer.Check(Condition: Boolean; const ErrMsg: String);
begin
  if not Condition then
    raise EMoleculeRenderer.Create(ErrMsg);
end;

procedure TMoleculeRenderer.ProcessRenderInfo(const RenderInfo: String;
  Rect: TRect; const ZoomFactor: Double; ACanvas: TCanvas; WhatToDo: Integer);
var
  FromPos, ToPos: Integer;
  TextHeight: Integer;
  TextOffsetY: Integer;
  CoordToPixelFactor: Double;
  // Local clip region to speed up rendering.
  ClipRegion: TRect;

  procedure NextDelim(const Delims: String; MustExist: Boolean);
  begin
    while (ToPos <= Length(RenderInfo)) and (not IsDelimiter(Delims, RenderInfo, ToPos)) do
      inc(ToPos);
    if MustExist then
      Check(ToPos <= Length(RenderInfo), 'Missing "' + Delims + '".');
  end;

  function NextInt: Integer;
  begin
    ToPos := FromPos;
    NextDelim(MOL_RNDR_DELIM, true);
    Check(ToPos <> FromPos, 'Missing value.');
    Result := StrToInt(copy(RenderInfo, FromPos, ToPos - FromPos));
    FromPos := ToPos + 1;
  end;

  function LastInt: Integer;
  begin
    ToPos := FromPos;
    NextDelim(MOL_RNDR_TAG_START, true);
    Check(ToPos <> FromPos, 'Missing value.');
    Result := StrToInt(copy(RenderInfo, FromPos, ToPos - FromPos));
    FromPos := ToPos;
  end;

  function NextOrLastInt: Integer;
  begin
    ToPos := FromPos;
    NextDelim(MOL_RNDR_DELIM + MOL_RNDR_TAG_START, true);
    Check(ToPos <> FromPos, 'Missing value.');
    Result := StrToInt(copy(RenderInfo, FromPos, ToPos - FromPos));
    FromPos := ToPos;
    if RenderInfo[FromPos] = MOL_RNDR_DELIM then
      inc(FromPos);
  end;

  function LastString: String;
  begin
    ToPos := FromPos;
    NextDelim(MOL_RNDR_TAG_START, true);
    Check(ToPos <> FromPos, 'Missing string value.');
    Result := Copy(RenderInfo, FromPos, ToPos - FromPos);
    FromPos := ToPos;
  end;

  procedure ReadHeader;
  begin
    // Valid render info at all ?
    Check(
      copy(RenderInfo, 1, Length(MOL_RNDR_TAG_INFO_BEGIN)) = MOL_RNDR_TAG_INFO_BEGIN,
      'Rendering info must start with header "' + MOL_RNDR_TAG_INFO_BEGIN + '".'
    );
    Check(
      copy(RenderInfo, Length(MOL_RNDR_TAG_INFO_BEGIN) + 1, 3) = MOL_RNDR_TAG_BOUNDBOX,
      'Rendering header must be followed by a bounding box "' + MOL_RNDR_TAG_BOUNDBOX + '".'
    );
    Check(
      copy(RenderInfo, Length(RenderInfo) - 2, 3) = MOL_RNDR_TAG_INFO_END,
      'Rendering info must end with an End-of-Data tag "' + MOL_RNDR_TAG_INFO_END + '".'
    );

    // Read bounding box values.
    FromPos := Length(MOL_RNDR_TAG_INFO_BEGIN) + Length(MOL_RNDR_TAG_BOUNDBOX) + 1;
    try
      CoordScaling := NextInt;
      XMax := NextInt;
      YMax := NextOrLastInt;
    except
      on E: Exception do
        raise EMoleculeRenderer.CreateFmt('Error reading bounding box: %s', [E.Message]);
    end;
  end;

  procedure CalcBoundingBox;
  begin
    BoxWidth  := Round(XMax * CoordToPixelFactor);
    BoxHeight := Round(YMax * CoordToPixelFactor);
  end;

  procedure RenderFrame;
  begin
    ACanvas.Pen.Style := psSolid;
    ACanvas.Pen.Color := clBlack;

    ACanvas.MoveTo(Rect.Left, Rect.Top);
    ACanvas.LineTo(Rect.Left + BoxWidth, Rect.Top);
    ACanvas.LineTo(Rect.Left + BoxWidth, Rect.Top + BoxHeight);
    ACanvas.LineTo(Rect.Left, Rect.Top + BoxHeight);
    ACanvas.LineTo(Rect.Left, Rect.Top);
  end;

  procedure RenderBond;
  var
    x1: Integer;
    y1: Integer;
    x2: Integer;
    y2: Integer;
    BondData: String;
    Cardinality: Char;
    ExtraInfo: Char;
    //
    SavedPenColor: TColor;
    SavedFontColor: TColor;
    SavedBrushColor: TColor;
    SavedBrushStyle: TBrushStyle;
    // For stereo-bond rendering.
    OrthoX: Double;
    OrthoY: Double;
    x1a: Double;
    y1a: Double;
    x2a: Double;
    y2a: Double;
    // For stereo-bond rendering.
    Triangle: array[0..2] of TPoint;
    // For APO rendering.
    ApoLabel: String;
    ApoLabelX, ApoLabelY: Integer;

    procedure DrawStereoWedge;
    begin
      OrthoX := -(y2 - y1) * 0.1;
      OrthoY :=  (x2 - x1) * 0.1;
      Triangle[0] := Point(Rect.Left + round(x1 * CoordToPixelFactor), Rect.Top + round(y1 * CoordToPixelFactor));
      Triangle[1] := Point(Rect.Left + round((x2 - OrthoX) * CoordToPixelFactor), Rect.Top + round((y2 - OrthoY) * CoordToPixelFactor));
      Triangle[2] := Point(Rect.Left + round((x2 + OrthoX) * CoordToPixelFactor), Rect.Top + round((y2 + OrthoY) * CoordToPixelFactor));
      ACanvas.Polygon(Triangle);
    end;

  begin
    x1 := NextInt;
    y1 := NextInt;
    x2 := NextInt;
    y2 := NextInt;
    BondData := LastString;
    if Length(BondData) <> 2 then
      raise EMoleculeRenderer.CreateFmt('Invalid bond info "%s". Expected two characters.', [BondData]);
    Cardinality := BondData[1];
    ExtraInfo   := BondData[2];

    ACanvas.Pen.Style := psSolid;

    if ( (x1 < ClipRegion.Left) or (x1 > ClipRegion.Right) or (y1 < ClipRegion.Top) or (y1 > ClipRegion.Bottom) ) and
       ( (x2 < ClipRegion.Left) or (x2 > ClipRegion.Right) or (y2 < ClipRegion.Top) or (y2 > ClipRegion.Bottom) ) then
      Exit;

    case Cardinality of

    // APO bond. ExtraInfo contains APO number.
    MOL_RNDR_BOND_CARDINALITY_APO: begin
      SavedPenColor   := ACanvas.Pen.Color;
      SavedFontColor  := ACanvas.Font.Color;
      SavedBrushStyle := ACanvas.Brush.Style;
      ACanvas.Pen.Color   := clRed;
      ACanvas.Font.Color  := clRed;
      ACanvas.Brush.Style := bsClear;
      ACanvas.Pen.Width   := 2 * ACanvas.Pen.Width;
      try
        ApoLabelX := Rect.Left + round(x2 * CoordToPixelFactor);
        ApoLabelY := Rect.Top + round(y2 * CoordToPixelFactor);
        ACanvas.MoveTo(Rect.Left + round(x1 * CoordToPixelFactor), Rect.Top + round(y1 * CoordToPixelFactor));
        ACanvas.LineTo(ApoLabelX, ApoLabelY);
        ApoLabelY := ApoLabelY + TextOffsetY;
        case ExtraInfo of
        '1': begin
               ApoLabel := '1*';
               ACanvas.TextOut(ApoLabelX - Round(ACanvas.TextWidth(ApoLabel) * 1.2), ApoLabelY, ApoLabel);
             end;
        '2': begin
               ApoLabel := '*2';
               ACanvas.TextOut(ApoLabelX + Round(ACanvas.TextWidth(ApoLabel) * 0.2), ApoLabelY, ApoLabel);
             end
        else
          raise EMoleculeRenderer.CreateFmt('Invalid APO number (%d).', [ExtraInfo]);
        end;
      finally
        ACanvas.Pen.Color   := SavedPenColor;
        ACanvas.Pen.Width   := ACanvas.Pen.Width div 2;
        ACanvas.Font.Color  := SavedFontColor;
        ACanvas.Brush.Style := SavedBrushStyle;
      end;
    end;

    // Single bond. ExtraInfo contains stereo-information.
    '1': begin
      case ExtraInfo of
      MOL_RNDR_STEREO_NONE: begin
        ACanvas.MoveTo(Rect.Left + round(x1 * CoordToPixelFactor), Rect.Top + round(y1 * CoordToPixelFactor));
        ACanvas.LineTo(Rect.Left + round(x2 * CoordToPixelFactor), Rect.Top + round(y2 * CoordToPixelFactor));
      end;
      MOL_RNDR_STEREO_UP: begin
        SavedBrushColor := ACanvas.Brush.Color;
        ACanvas.Brush.Color := clBlack;
        try
          DrawStereoWedge;
        finally
          ACanvas.Brush.Color := SavedBrushColor;
        end;
      end;
      MOL_RNDR_STEREO_DOWN: begin
        SavedBrushColor := ACanvas.Brush.Color;
        SavedPenColor   := ACanvas.Pen.Color;
        ACanvas.Brush.Color := clLtGray;
        ACanvas.Pen.Color   := clLtGray;
        try
          DrawStereoWedge;
        finally
          ACanvas.Brush.Color := SavedBrushColor;
          ACanvas.Pen.Color   := SavedPenColor;
        end;
      end
      else
        raise EMoleculeRenderer.CreateFmt('Invalid bond stereo configuration (%d).', [ExtraInfo]);
      end;
    end;

    // Double bond. ExtraInfo contains information on how to display aromatic bonds in rings.
    '2': begin
      OrthoX := -(y2 - y1) * 0.1;
      OrthoY :=  (x2 - x1) * 0.1;
      if ExtraInfo <> MOL_RNDR_DBLBOND_SYMMETRICAL then begin
        OrthoX := OrthoX * 1.6;
        OrthoY := OrthoY * 1.6;
        x1a := x1 + OrthoY;
        y1a := y1 - OrthoX;
        x2a := x2 - OrthoY;
        y2a := y2 + OrthoX;
      end
      else begin
        // This block to avoid compiler warnings - and possible future maintenance problems.
        x1a := 0;
        y1a := 0;
        x2a := 0;
        y2a := 0;
      end;
      case ExtraInfo of
        MOL_RNDR_DBLBOND_SYMMETRICAL: begin
          ACanvas.MoveTo(Rect.Left + round((x1 - OrthoX) * CoordToPixelFactor), Rect.Top + round((y1 - OrthoY) * CoordToPixelFactor));
          ACanvas.LineTo(Rect.Left + round((x2 - OrthoX) * CoordToPixelFactor), Rect.Top + round((y2 - OrthoY) * CoordToPixelFactor));
          ACanvas.MoveTo(Rect.Left + round((x1 + OrthoX) * CoordToPixelFactor), Rect.Top + round((y1 + OrthoY) * CoordToPixelFactor));
          ACanvas.LineTo(Rect.Left + round((x2 + OrthoX) * CoordToPixelFactor), Rect.Top + round((y2 + OrthoY) * CoordToPixelFactor));
        end;
        MOL_RNDR_DBLBOND_LEFT: begin
          ACanvas.MoveTo(Rect.Left + round(x1 * CoordToPixelFactor), Rect.Top + round(y1 * CoordToPixelFactor));
          ACanvas.LineTo(Rect.Left + round(x2 * CoordToPixelFactor), Rect.Top + round(y2 * CoordToPixelFactor));
          ACanvas.MoveTo(Rect.Left + round((x1a - OrthoX) * CoordToPixelFactor), Rect.Top + round((y1a - OrthoY) * CoordToPixelFactor));
          ACanvas.LineTo(Rect.Left + round((x2a - OrthoX) * CoordToPixelFactor), Rect.Top + round((y2a - OrthoY) * CoordToPixelFactor));
          {$ifdef DEBUG_GRAPHICS} // Show bond origin with small circle to check "left" orientation.
          ACanvas.Ellipse(
            Rect.Left + round((x1a - OrthoX) * CoordToPixelFactor) - 2,
            Rect.Top + round((y1a - OrthoY) * CoordToPixelFactor) - 2,
            Rect.Left + round((x1a - OrthoX) * CoordToPixelFactor) + 2,
            Rect.Top + round((y1a - OrthoY) * CoordToPixelFactor) + 2
          );
          {$endif}
        end;
        MOL_RNDR_DBLBOND_RIGHT: begin
          ACanvas.MoveTo(Rect.Left + round(x1 * CoordToPixelFactor), Rect.Top + round(y1 * CoordToPixelFactor));
          ACanvas.LineTo(Rect.Left + round(x2 * CoordToPixelFactor), Rect.Top + round(y2 * CoordToPixelFactor));
          ACanvas.MoveTo(Rect.Left + round((x1a + OrthoX) * CoordToPixelFactor), Rect.Top + round((y1a + OrthoY) * CoordToPixelFactor));
          ACanvas.LineTo(Rect.Left + round((x2a + OrthoX) * CoordToPixelFactor), Rect.Top + round((y2a + OrthoY) * CoordToPixelFactor));
          {$ifdef DEBUG_GRAPHICS} // Show bond origin with large circle to check "right" orientation.
          ACanvas.Ellipse(
            Rect.Left + round((x1a + OrthoX) * CoordToPixelFactor) - 4,
            Rect.Top + round((y1a + OrthoY) * CoordToPixelFactor) - 4,
            Rect.Left + round((x1a + OrthoX) * CoordToPixelFactor) + 4,
            Rect.Top + round((y1a + OrthoY) * CoordToPixelFactor) + 4
          );
          {$endif}
        end;
        else
          raise EMoleculeRenderer.CreateFmt('Invalid double-bond extra info (%d).', [ExtraInfo]);
      end;
    end;

    // Triple bond.
    '3': begin
      { Triple bonds need more spacing than double bonds so the middle line can be seen. }
      OrthoX := -(y2 - y1) * 0.16;
      OrthoY :=  (x2 - x1) * 0.16;
      // Double bond
      ACanvas.MoveTo(Rect.Left + round((x1 - OrthoX) * CoordToPixelFactor), Rect.Top + round((y1 - OrthoY) * CoordToPixelFactor));
      ACanvas.LineTo(Rect.Left + round((x2 - OrthoX) * CoordToPixelFactor), Rect.Top + round((y2 - OrthoY) * CoordToPixelFactor));
      ACanvas.MoveTo(Rect.Left + round((x1 + OrthoX) * CoordToPixelFactor), Rect.Top + round((y1 + OrthoY) * CoordToPixelFactor));
      ACanvas.LineTo(Rect.Left + round((x2 + OrthoX) * CoordToPixelFactor), Rect.Top + round((y2 + OrthoY) * CoordToPixelFactor));
      // plus the ordinary line in the middle.
      ACanvas.MoveTo(Rect.Left + round(x1 * CoordToPixelFactor), Rect.Top + round(y1 * CoordToPixelFactor));
      ACanvas.LineTo(Rect.Left + round(x2 * CoordToPixelFactor), Rect.Top + round(y2 * CoordToPixelFactor));
    end

    else
      raise EMoleculeRenderer.CreateFmt('Bond cardinality out of range (%d).', [Cardinality]);
    end;
  end;

  procedure CalculateResidueFontSizeDelta;
  begin
    if fOneLetterResidues then
      fResidueFontSizeDelta := 1
    else begin
      fResidueFontSizeDelta := -(ACanvas.Font.Size div 4);
    end;
  end;

  procedure RenderAtom;
  var
    i: Integer;
    x: Integer;
    y: Integer;
    Symbol: String;
    ExtraInfo: String;
    ChargeInfo: String;
    ChainIndex: Integer;
    TextWidth: Integer;
    TextOffsetX: Integer;
    IsPseudoAtom: Boolean;
    ResidueCode: Integer;
  begin
    x := NextInt;
    y := NextInt;
    Symbol := LastString;

    if (x < ClipRegion.Left) or (x > ClipRegion.Right) or (y < ClipRegion.Top) or (y > ClipRegion.Bottom) then
      Exit;

    ChainIndex := 0;

    i := Pos(',', Symbol);
    if i > 0 then begin
      ExtraInfo := Copy(Symbol, i + 1, MaxInt);
      Delete(Symbol, i, MaxInt);
    end
    else
      ExtraInfo := '';

    IsPseudoAtom := (Length(Symbol) = 3);
    if IsPseudoAtom then begin
      ResidueCode  := _ThreeLetterCodes.IndexOf(Symbol);
      IsPseudoAtom := ResidueCode >= 0;
      if IsPseudoAtom and fOneLetterResidues then
        Symbol := _OneLetterCodes[ResidueCode + 1];

      if ExtraInfo <> '' then begin
        ChainIndex := StrToInt(ExtraInfo) - 1;
        if ChainIndex <= 0 then
          ChainIndex := 0;
      end;
    end;

    if IsPseudoAtom then begin
      if not fHasCalculatedResidueFontSize then
        CalculateResidueFontSizeDelta;
      ACanvas.Font.Style := ACanvas.Font.Style + [fsBold];
      if ACanvas.Font.Size >= 4 then begin
        ACanvas.Font.Size  := ACanvas.Font.Size + fResidueFontSizeDelta;
        TextHeight  := ACanvas.TextHeight('W');
        TextOffsetY := -TextHeight div 2;
      end;
    end;

    TextWidth   := ACanvas.TextWidth(Symbol);
    TextOffsetX := TextWidth div 2;

    if (Symbol = 'R#') or (Symbol = 'R1') then
      ACanvas.Font.Color := clRed
    else
      ACanvas.Font.Color := clBlack;

    x := Rect.Left + Round(x * CoordToPixelFactor);
    y := Rect.Top  + Round(y * CoordToPixelFactor);

    // Clear text background with .Rectangle() to work around OOo issue 107453.
    // Otherwise bonds will show through text background even though text is
    // rendering with Brush.Style = bsSolid.
    // Only relevant when using the renderer to produce a TMetafile image via
    // a TMetafileCanvas.
    ACanvas.Pen.Style := psClear;
    ACanvas.Rectangle(x - TextOffsetX, y + TextOffsetY, x - TextOffsetX + TextWidth, y + TextOffsetY + TextHeight);
    // Workaround end.

    if IsPseudoAtom then begin
      ACanvas.Pen.Color := clBlack;
      ACanvas.Pen.Style := psSolid;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Brush.Color :=
        CHAIN_BACKGND_COLORS[ChainIndex mod NUM_CHAIN_BACKGND_COLORS];
      // +1 seems to compensate nicely for the slightly off-center labels (visual artefact due to the italics font).
      ACanvas.Ellipse(x - fResidueCircleRadius + 1, y - fResidueCircleRadius, x + fResidueCircleRadius + 1, y + fResidueCircleRadius);
      ACanvas.Brush.Style := bsClear;
    end;
    ACanvas.TextOut(x - TextOffsetX, y + TextOffsetY, Symbol);
    if IsPseudoAtom then begin
      ACanvas.Font.Style := ACanvas.Font.Style - [fsBold];
      if ACanvas.Font.Size >= 2 then begin
        ACanvas.Font.Size  := ACanvas.Font.Size - fResidueFontSizeDelta;
        TextHeight  := ACanvas.TextHeight('W');
        TextOffsetY := -TextHeight div 2;
      end;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Brush.Color := clWindow;
    end;

    // Charges ?
    if ExtraInfo <> '' then begin
      if (ExtraInfo[1] = '+') or (ExtraInfo[1] = '-') then begin
        ChargeInfo := ExtraInfo;
        i := 2;
        while (ChargeInfo[i] in ['0'..'9']) and (i <= Length(ChargeInfo)) do
          inc(i);
        if i < Length(ChargeInfo) then
          Delete(ChargeInfo, i, MaxInt);
        if (Length(ChargeInfo) = 2) and (ChargeInfo[2] = '1') then
          Delete(ChargeInfo, 2, 1)
        else
          ChargeInfo := Copy(ChargeInfo, 2, MaxInt) + ChargeInfo[1];
        ACanvas.TextOut(x + TextOffsetX, y + 2 * TextOffsetY, ChargeInfo);
      end;
    end;
  end;

  procedure RenderLabel;
  var
    x: Integer;
    y: Integer;
    Alignment: Char;
    TheLabel: String;
    TextWidth: Integer;
  begin
    x := NextInt;
    y := NextInt;
    TheLabel := LastString;
    if (x < ClipRegion.Left) or (x > ClipRegion.Right) or (y < ClipRegion.Top) or (y > ClipRegion.Bottom) then
      Exit;

    if Length(TheLabel) < 2 then
      raise EMoleculeRenderer.CreateFmt('Alignment + label must be more than 2 characters: "%s".', [TheLabel]);
    if TheLabel[2] <> MOL_RNDR_DELIM then
      raise EMoleculeRenderer.CreateFmt('Alignment and label are not separated by the expected comma-delimiter: "%s".', [TheLabel]);
    Alignment := TheLabel[1];
    Delete(TheLabel, 1, 2);

    TheLabel := StringReplace(TheLabel, '%3A', ':', [rfReplaceAll]);
    TheLabel := StringReplace(TheLabel, '%2C', ',', [rfReplaceAll]);
    TheLabel := StringReplace(TheLabel, '%25', '%', [rfReplaceAll]);

    TextWidth   := ACanvas.TextWidth(TheLabel);

    x := Rect.Left + Round(x * CoordToPixelFactor);
    y := Rect.Top  + Round(y * CoordToPixelFactor);

    if Alignment = MOL_RNDR_LABEL_ALIGN_CENTER then
      x := x - TextWidth div 2
    else if Alignment = MOL_RNDR_LABEL_ALIGN_LEFT then
      x := x
    else if Alignment = MOL_RNDR_LABEL_ALIGN_RIGHT then
      x := x - TextWidth
    else
      raise EMoleculeRenderer.CreateFmt('Invalid label alignment "%s".', [Alignment]);

    ACanvas.TextOut(x, y + TextOffsetY, TheLabel);
  end;

const
  COORD_TO_PIXEL_SCALING = 60; // Factor to use for a zoom of 1.0 when coordinate scale factor is 1000.
var
  Tag: String;

  procedure NextTag;
  begin
    Tag     := copy(RenderInfo, FromPos, 3);
    FromPos := FromPos + 3;
  end;

var
  SavedFontSize: Integer;
begin
  if RenderInfo = '' then
    Exit;

  if WhatToDo = DO_CALCBB then begin
    ReadHeader;
    CoordToPixelFactor := ZoomFactor / COORD_TO_PIXEL_SCALING * (1000 / CoordScaling);
    CalcBoundingBox;
    Exit;
  end
  else
    if WhatToDo <> DO_RENDER then
      raise EMoleculeRenderer.CreateFmt('Internal bug: Invalid request sent to ProcessRenderInfo (%d).', [WhatToDo]);

  ACanvas.Font.Style  := [];
  ACanvas.Brush.Style := bsSolid;
  try
    TextHeight  := ACanvas.TextHeight('W');
    TextOffsetY := -TextHeight div 2;

    SavedFontSize := ACanvas.Font.Size;

    if OneLetterResidues then
      fResidueSymbolTemplate := 'W' + '  '
    else
      fResidueSymbolTemplate := 'Ala' + '   ';

    fHasCalculatedResidueFontSize := false;
    fResidueFontSizeDelta := 0;

    ReadHeader;
    CoordToPixelFactor := ZoomFactor / COORD_TO_PIXEL_SCALING * (1000 / CoordScaling);
    CalcBoundingBox;
    // Calculate local clip region boundaries. Pre-filtering bond and atom
    // coordinates instead of relying on GDI clipping speeds up rendering of
    // very large molecules (fewer relatively expensive graphics subsystem calls).
    if Abs(CoordToPixelFactor) > 0.000001 then begin
      ClipRegion.Left := Round( (-Rect.Left - ACanvas.TextWidth('W') * 8) / CoordToPixelFactor );
      ClipRegion.Top  := Round( (-Rect.Top - ACanvas.TextHeight('W') * 5) / CoordToPixelFactor );
      ClipRegion.Right  := ClipRegion.Left + Round( (Rect.Right - Rect.Left + ACanvas.TextWidth('W') * 16)  / CoordToPixelFactor );
      ClipRegion.Bottom := ClipRegion.Top  + Round( (Rect.Bottom - Rect.Top + ACanvas.TextHeight('W') * 10) / CoordToPixelFactor );
    end
    else begin
      ClipRegion.Left := 0;
      ClipRegion.Top  := 0;
      ClipRegion.Right  := XMax;
      ClipRegion.Bottom := YMax;
    end;

    fResidueCircleRadius := Round(0.45 * CoordScaling * CoordToPixelFactor);
//(** DEBUG:    RenderFrame;

    NextTag;
    try
      while Tag = MOL_RNDR_TAG_BOND do begin
        try
          RenderBond;
          NextTag;
        except
          on E: Exception do
            raise EMoleculeRenderer.CreateFmt('Error rendering bond: %s', [E.Message]);
        end;
      end;

      while Tag = MOL_RNDR_TAG_ATOM do begin
        try
          RenderAtom;
          NextTag;
        except
          on E: Exception do
            raise EMoleculeRenderer.CreateFmt('Error rendering atom: %s', [E.Message]);
        end;
      end;

      if Tag = MOL_RNDR_TAG_LABEL then begin
        ACanvas.Brush.Style := bsClear;
        ACanvas.Font.Size  := ACanvas.Font.Size - Min(2, ACanvas.Font.Size - 1);
        ACanvas.Font.Style := ACanvas.Font.Style + [fsBold];
        TextHeight   := ACanvas.TextHeight('W');
        TextOffsetY  := -TextHeight div 2;
      end;
      while Tag = MOL_RNDR_TAG_LABEL do begin
        try
          RenderLabel;
          NextTag;
        except
          on E: Exception do
            raise EMoleculeRenderer.CreateFmt('Error rendering label: %s', [E.Message]);
        end;
      end;
    finally
      ACanvas.Font.Size := SavedFontSize;
    end;

    if Tag <> MOL_RNDR_TAG_INFO_END then
      raise EMoleculeRenderer.CreateFmt('Unsupported render info tag "%s".', [Tag]);

  except
    on E: Exception do begin
      ACanvas.Pen.Color := clBlack;
      ACanvas.Pen.Style := psSolid;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Brush.Color := clWindow;
      ACanvas.TextOut(Rect.Left + 10, Rect.Top + 10, E.ClassName + ': ' + E.Message);
    end;
  end;
end;

function TMoleculeRenderer.BoundingBox(const RenderInfo: String; const ZoomFactor: Double;
  ACanvas: TCanvas): TRect;
begin
  try
    ProcessRenderInfo(RenderInfo, Rect(0, 0, 100, 100), ZoomFactor, ACanvas, DO_CALCBB);
    Result.Left := 0;
    Result.Top  := 0;
    Result.Right  := BoxWidth;
    Result.Bottom := BoxHeight;
  except
    Result.Left := 0;
    Result.Top  := 0;
    Result.Right  := 1;
    Result.Bottom := 1;
  end;
end;

procedure TMoleculeRenderer.Render(const RenderInfo: String; Rect: TRect;
  const ZoomFactor: Double; ACanvas: TCanvas);
begin
  ProcessRenderInfo(RenderInfo, Rect, ZoomFactor, ACanvas, DO_RENDER);
end;

initialization
  _ThreeLetterCodes := TStringList.Create;
  _ThreeLetterCodes.Add('Ala');
  _ThreeLetterCodes.Add('Arg');
  _ThreeLetterCodes.Add('Asn');
  _ThreeLetterCodes.Add('Asp');
  _ThreeLetterCodes.Add('Cys');
  _ThreeLetterCodes.Add('Glu');
  _ThreeLetterCodes.Add('Gln');
  _ThreeLetterCodes.Add('Gly');
  _ThreeLetterCodes.Add('His');
  _ThreeLetterCodes.Add('Ile');
  _ThreeLetterCodes.Add('Leu');
  _ThreeLetterCodes.Add('Lys');
  _ThreeLetterCodes.Add('Met');
  _ThreeLetterCodes.Add('Phe');
  _ThreeLetterCodes.Add('Pro');
  _ThreeLetterCodes.Add('Ser');
  _ThreeLetterCodes.Add('Thr');
  _ThreeLetterCodes.Add('Trp');
  _ThreeLetterCodes.Add('Tyr');
  _ThreeLetterCodes.Add('Val');
  _ThreeLetterCodes.Add('Sec');

  _OneLetterCodes := 'ARNDCEQGHILKMFPSTWYVU';
finalization
  _ThreeLetterCodes.Free;
end.

