/**
  2D molecule renderer that parses molecule rendering info V2 produced by
  Proteax and generates corresponding graphical output on a
  System.Drawing.Graphics device.

(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This C# source code unit (MoleculeRenderer.cs) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/

/**
 * NOTE 2010-12-14; TODO-items:
 * 
 1) Label placement is slightly off X-wise - too large a width being reported ?
 2) Line aliasing would be rather nice - current display is somewhat jagged.
 * 
**/

using System;
using System.Drawing;

namespace Biochemfusion
{

	class MoleculeRenderer
	{
		const char MOL_RNDR_DELIM = ',';		// Value delimiter.
		const char MOL_RNDR_TAG_START = ':';	// Tag start.

		const string MOL_RNDR_TAG_BOUNDBOX = "BB";	// Bounding box.
		const string MOL_RNDR_TAG_BOND = "BD";		// Bond (incl. APO bond).
		const string MOL_RNDR_TAG_ATOM = "AT";		// Atom.
		const string MOL_RNDR_TAG_LABEL = "LB";		// Label.

		const char MOL_RNDR_BOND_CARDINALITY_APO = 'a';

		const char MOL_RNDR_STEREO_NONE = '.';
		const char MOL_RNDR_STEREO_UP   = 'u';
		const char MOL_RNDR_STEREO_DOWN = 'd';

		const char MOL_RNDR_DBLBOND_SYMMETRICAL = 's';
		const char MOL_RNDR_DBLBOND_LEFT        = 'l';
		const char MOL_RNDR_DBLBOND_RIGHT       = 'r';

		const char MOL_RNDR_LABEL_ALIGN_CENTER	= 'c';
		const char MOL_RNDR_LABEL_ALIGN_LEFT	= 'l';
		const char MOL_RNDR_LABEL_ALIGN_RIGHT	= 'r';

		const string MOL_RNDR_TAG_INFO_BEGIN = "BCF_MOL_RNDR_INFO_V2";
		const string MOL_RNDR_TAG_INFO_END = "ED";		// End of Data.

		private readonly /* "const" */ string[] _threeLetterCodes = new string[] {
			"Ala", "Arg", "Asn", "Asp", "Cys", "Glu",
			"Gln", "Gly", "His", "Ile", "Leu", "Lys",
			"Met", "Phe", "Pro", "Ser", "Thr", "Trp",
			"Tyr", "Val", "Sec"
		};
		private const string _oneLetterCodes = "ARNDCEQGHILKMFPSTWYVU";

		private const int DO_RENDER = 0;
		private const int DO_CALCBB = 1;

		const int NUM_CHAIN_BACKGND_COLORS = 2;
		// Background colors use direct RGB values. None of the System.Drawing.Color constants
		// match Delphi's clMoneyGreen and clSkyBlue well.
		private readonly /* "const" */ Brush [] CHAIN_BACKGND_COLORS = new Brush[NUM_CHAIN_BACKGND_COLORS] {
			new SolidBrush(Color.FromArgb(192, 220, 192)),
			new SolidBrush(Color.FromArgb(166, 202, 240))
		};

		const int COORD_TO_PIXEL_SCALING = 60; // Factor to use for a zoom of 1.0 when coordinate scale factor is 1000.
		float xMax;
		float yMax;
		float coordScaling;
		int boxWidth;
		int boxHeight;

		Rectangle rect;
		Graphics canvas;

		float atomTextWidth;
		float atomTextHeight;

		float labelTextHeight;
		float labelTextOffsetY;

		Rectangle clipRegion = new Rectangle();

		bool _oneLetterResidues;
		float residueCircleRadius;

		const string FONT_NAME = "Arial";
		const int DEFAULT_ATOM_FONT_SIZE = 6;
		const int ERROR_MESSAGE_FONT_SIZE = 8;
		// The default font provides a measure of text size at 100% zoom.
		Font defaultFont = new Font(FONT_NAME, DEFAULT_ATOM_FONT_SIZE);
		Font atomFont;
		Font residueFont;
		Font labelFont;

		Brush atomBrush = new SolidBrush(Color.Black);
		Brush residueBrush = new SolidBrush(Color.Black);
		Brush labelBrush = new SolidBrush(Color.Black);
		Brush apoLabelBrush = new SolidBrush(Color.Red);
		Brush rgroupAtomBrush = new SolidBrush(Color.Red);

		Pen bondPen = new Pen(new SolidBrush(Color.Black));
		Pen apoBondPen = new Pen(new SolidBrush(Color.Red), 2);
		Brush upBondBrush = new SolidBrush(Color.Black);
		Brush downBondBrush = new SolidBrush(Color.LightGray);

		Brush backgroundBrush = new SolidBrush(Color.White);

		float coordToPixelFactor;

		int tokenIdx;
		string currentToken;
		string tag;

		private void check(bool condition, string errMsg)
		{
			if (!condition)
				throw new Exception(errMsg);
		}

		private void readHeader(string[] tokens)
		{
		    // Valid render info at all ?
			// Note: tokens[0] is blank since render info starts with a delimiter.
			check(
				tokens.Length > 1 && tokens[1] == MOL_RNDR_TAG_INFO_BEGIN,
				"Rendering info must start with header '" + MOL_RNDR_TAG_INFO_BEGIN + "'."
			);
			check(
				tokens.Length > 2 && tokens[2].Substring(0, 2) == MOL_RNDR_TAG_BOUNDBOX,
				"Rendering header must be followed by a bounding box '" + MOL_RNDR_TAG_BOUNDBOX + "'."
			);
			check(
				tokens[tokens.Length - 1] == MOL_RNDR_TAG_INFO_END,
				"Rendering info must end with an End-of-Data tag '" + MOL_RNDR_TAG_INFO_END + "'."
			);

			// Read bounding box values.
			string[] values = tokens[2].Split(MOL_RNDR_DELIM);
			check(values.Length == 3, "Invalid bounding box - expected 5 values.");
			try {
			  coordScaling = int.Parse(values[0].Remove(0, 2));
			  xMax         = int.Parse(values[1]);
			  yMax         = int.Parse(values[2]);
			} catch (Exception e) {
				throw new Exception ("Error reading bounding box: " + e.ToString());
			}
		}

		private void calcBoundingBox()
		{
			boxWidth  = (int) Math.Round(xMax * coordToPixelFactor);
			boxHeight = (int) Math.Round(yMax * coordToPixelFactor);
		}

		private void renderFrame()
		{
			Pen pen = new Pen(new SolidBrush(Color.Black));
			canvas.DrawLine(pen, new Point(rect.X, rect.Y), new Point(rect.X + boxWidth, rect.Y));
			canvas.DrawLine(pen, new Point(rect.X + boxWidth, rect.Y), new Point(rect.X + boxWidth, rect.Y + boxHeight));
			canvas.DrawLine(pen, new Point(rect.X, rect.Y + boxHeight), new Point(rect.X + boxWidth, rect.Y + boxHeight));
			canvas.DrawLine(pen, new Point(rect.X, rect.Y + boxHeight), new Point(rect.X, rect.Y));
		}

		private void nextToken(string[] tokens)
		{
			tokenIdx++;
			currentToken = tokens[tokenIdx];
			tag = currentToken.Substring(0, 2);
			currentToken = currentToken.Remove(0, 2);
		}

		private void drawStereoWedge(float x1, float y1, float x2, float y2, Brush brush)
		{
			float orthoX = -(y2 - y1) * (float) 0.1;
			float orthoY =  (x2 - x1) * (float) 0.1;

			canvas.FillPolygon(brush, new PointF[] {
				new PointF(rect.X + x1 * coordToPixelFactor, rect.Y + y1 * coordToPixelFactor),
				new PointF(rect.X + (x2 - orthoX) * coordToPixelFactor, rect.Y + (y2 - orthoY) * coordToPixelFactor),
				new PointF(rect.X + (x2 + orthoX) * coordToPixelFactor, rect.Y + (y2 + orthoY) * coordToPixelFactor)
			});
		}

		private void renderBond()
		{
			string[] values = currentToken.Split(MOL_RNDR_DELIM);
			if (values.Length != 5)
				throw new Exception ("Invalid bond token '" + currentToken + "' - must consist of 5 values.");

			float x1 = int.Parse(values[0]);
			float y1 = int.Parse(values[1]);
			float x2 = int.Parse(values[2]);
			float y2 = int.Parse(values[3]);

			if ( ( x1 < clipRegion.X || x1 > clipRegion.X + clipRegion.Width || y1 < clipRegion.Y || y1 > clipRegion.Y + clipRegion.Height ) &&
			     ( x2 < clipRegion.X || x2 > clipRegion.X + clipRegion.Width || y2 < clipRegion.Y || y2 > clipRegion.Y + clipRegion.Height ) )
				return;			

			string bondData = values[4];
			if (bondData.Length != 2)
				throw new Exception ("Invalid bond info '" + bondData + "'. Expected two characters.");
			char cardinality = bondData[0];
			char extraInfo   = bondData[1];

			if (cardinality == MOL_RNDR_BOND_CARDINALITY_APO) {
				// APO bond. ExtraInfo contains APO number.
				float apoLabelX = rect.X + x2 * coordToPixelFactor;
				float apoLabelY = rect.Y + y2 * coordToPixelFactor;
				canvas.DrawLine(
					apoBondPen,
					rect.X + x1 * coordToPixelFactor, rect.Y + y1 * coordToPixelFactor,
					apoLabelX, apoLabelY
				);
				apoLabelY -= canvas.MeasureString("1*", atomFont).Height / 2;

				if (extraInfo == '1')
					canvas.DrawString("1*", atomFont, apoLabelBrush, apoLabelX - canvas.MeasureString("1*", atomFont).Width, apoLabelY);
				else if (extraInfo == '2')
					canvas.DrawString("*2", atomFont, apoLabelBrush, apoLabelX, apoLabelY);
				else
					throw new Exception ("Invalid APO number (" + extraInfo + ").");
			}
			else if (cardinality == '1') {
				// Single bond. ExtraInfo contains stereo-information.
				if (extraInfo == MOL_RNDR_STEREO_NONE)
					canvas.DrawLine(
						bondPen,
						rect.X + x1 * coordToPixelFactor, rect.Y + y1 * coordToPixelFactor,
						rect.X + x2 * coordToPixelFactor, rect.Y + y2 * coordToPixelFactor
					);
				else if (extraInfo == MOL_RNDR_STEREO_UP)
					drawStereoWedge(x1, y1, x2, y2, upBondBrush);
				else if (extraInfo == MOL_RNDR_STEREO_DOWN)
					drawStereoWedge(x1, y1, x2, y2, downBondBrush);
				else
					throw new Exception ("Invalid bond stereo configuration (" + extraInfo + ").");
			}
			else if (cardinality == '2') {
				// Double bond. ExtraInfo contains information on how to display aromatic bonds in rings.
				float orthoX = -(y2 - y1) * (float) 0.1;
				float orthoY =  (x2 - x1) * (float) 0.1;

				float x1a = 0;
				float y1a = 0;
				float x2a = 0;
				float y2a = 0;

				if (extraInfo != MOL_RNDR_DBLBOND_SYMMETRICAL) {
					orthoX = orthoX * (float) 1.6;
					orthoY = orthoY * (float) 1.6;
					x1a = x1 + orthoY;
					y1a = y1 - orthoX;
					x2a = x2 - orthoY;
					y2a = y2 + orthoX;
				}

				if (extraInfo == MOL_RNDR_DBLBOND_SYMMETRICAL) {
					canvas.DrawLine(
						bondPen,
						rect.X + (x1 - orthoX) * coordToPixelFactor, rect.Y + (y1 - orthoY) * coordToPixelFactor,
						rect.X + (x2 - orthoX) * coordToPixelFactor, rect.Y + (y2 - orthoY) * coordToPixelFactor
					);
					canvas.DrawLine(
						bondPen,
						rect.X + (x1 + orthoX) * coordToPixelFactor, rect.Y + (y1 + orthoY) * coordToPixelFactor,
						rect.X + (x2 + orthoX) * coordToPixelFactor, rect.Y + (y2 + orthoY) * coordToPixelFactor
					);
				}
				else if (extraInfo == MOL_RNDR_DBLBOND_LEFT)
				{
					canvas.DrawLine(
						bondPen,
						rect.X + x1 * coordToPixelFactor, rect.Y + y1 * coordToPixelFactor,
						rect.X + x2 * coordToPixelFactor, rect.Y + y2 * coordToPixelFactor
					);
					canvas.DrawLine(
						bondPen,
						rect.X + (x1a - orthoX) * coordToPixelFactor, rect.Y + (y1a - orthoY) * coordToPixelFactor,
						rect.X + (x2a - orthoX) * coordToPixelFactor, rect.Y + (y2a - orthoY) * coordToPixelFactor
					);
				}
				else if (extraInfo == MOL_RNDR_DBLBOND_RIGHT)
				{
					canvas.DrawLine(
						bondPen,
						rect.X + x1 * coordToPixelFactor, rect.Y + y1 * coordToPixelFactor,
						rect.X + x2 * coordToPixelFactor, rect.Y + y2 * coordToPixelFactor
					);
					canvas.DrawLine(
						bondPen,
						rect.X + (x1a + orthoX) * coordToPixelFactor, rect.Y + (y1a + orthoY) * coordToPixelFactor,
						rect.X + (x2a + orthoX) * coordToPixelFactor, rect.Y + (y2a + orthoY) * coordToPixelFactor
					);
				}
				else
					throw new Exception("Invalid double-bond extra info (" + extraInfo + ").");
			}
			else if (cardinality == '3') {
				// Triple bonds. They need more spacing than double bonds so the middle line can be seen.
				float orthoX = -(y2 - y1) * (float) 0.16;
				float orthoY =  (x2 - x1) * (float) 0.16;
				// Draw a double bond...
				canvas.DrawLine(
					bondPen,
					rect.X + (x1 - orthoX) * coordToPixelFactor, rect.Y + (y1 - orthoY) * coordToPixelFactor,
					rect.X + (x2 - orthoX) * coordToPixelFactor, rect.Y + (y2 - orthoY) * coordToPixelFactor
				);
				canvas.DrawLine(
					bondPen,
					rect.X + (x1 + orthoX) * coordToPixelFactor, rect.Y + (y1 + orthoY) * coordToPixelFactor,
					rect.X + (x2 + orthoX) * coordToPixelFactor, rect.Y + (y2 + orthoY) * coordToPixelFactor
				);
				// ... plus the ordinary line in the middle.
				canvas.DrawLine(
					bondPen,
					rect.X + x1 * coordToPixelFactor, rect.Y + y1 * coordToPixelFactor,
					rect.X + x2 * coordToPixelFactor, rect.Y + y2 * coordToPixelFactor
				);
			}
			else
				throw new Exception ("Bond cardinality out of range (" + cardinality + ").");
		}

		/**
		  Lookup a three-letter residue atom in the _threeLetterCodes array.
		  Returns the atom's corresponding index into _oneLetterCodes or -1 if not found.

		  Could use a Dictionary instead, but that requires .NET generics. A lookup into this small table
		  won't benefit much from using a Dictionary performance-wise and a Dictionary is harder to instantiate
		  so we'll go for the minimum requirements instead.
		**/
		private int oneLetterIndex(string atomSymbol)
		{
			int result = -1;
			if (atomSymbol.Length != 3)
				return result;

			for (int i = 0; i < _threeLetterCodes.Length; i++)
				if (_threeLetterCodes[i] == atomSymbol)
				{
					result = i;
					break;
				}

			return result;
		}

		private void renderAtom()
		{
			string[] values = currentToken.Split(MOL_RNDR_DELIM);
			if (values.Length != 3 && values.Length != 4)
				throw new Exception("Invalid atom token '" + currentToken + "' - must consist of 3 or 4 values.");

			float x = int.Parse(values[0]);
			float y = int.Parse(values[1]);
			string symbol = values[2];

			if ( x < clipRegion.X || x > clipRegion.X + clipRegion.Width || y < clipRegion.Y || y > clipRegion.Y + clipRegion.Height )
				return;

			Font textFont = atomFont;

			int chainIndex = 0;
			int residueIndex = oneLetterIndex(symbol);
			bool isPseudoAtom = residueIndex >= 0;

			if (isPseudoAtom) {
				if (_oneLetterResidues)
					symbol = new string( _oneLetterCodes[residueIndex], 1);
				if (values.Length == 4)
					chainIndex = int.Parse(values[3]) - 1;
				// Guard against invalid chain number value of "0".
				if (chainIndex < 0)
					chainIndex = 0;

				textFont = residueFont;
			}

			float textWidth   = canvas.MeasureString(symbol, textFont).Width;
			float textOffsetX = textWidth / 2;

			float atomTextOffsetY = - canvas.MeasureString("W", textFont).Height / 2;

			Brush atomColor = atomBrush;
			if (symbol == "R#" || symbol == "R1")
				atomColor = rgroupAtomBrush;

			x = rect.X + x * coordToPixelFactor;
			y = rect.Y + y * coordToPixelFactor;

			// Clear text background first.
			canvas.FillRectangle(
				backgroundBrush, x - textOffsetX, y + atomTextOffsetY, textWidth, atomTextHeight
			);

			if (isPseudoAtom) {
				Brush bubbleColor = CHAIN_BACKGND_COLORS[chainIndex % NUM_CHAIN_BACKGND_COLORS];
				canvas.FillEllipse(bubbleColor, x - residueCircleRadius, y - residueCircleRadius, residueCircleRadius * 2, residueCircleRadius * 2);
				canvas.DrawEllipse(bondPen, x - residueCircleRadius, y - residueCircleRadius, residueCircleRadius * 2, residueCircleRadius * 2);
			}

			canvas.DrawString(symbol, textFont, atomColor, x - textOffsetX, y + atomTextOffsetY);

			// Charges ?
			if (!isPseudoAtom && values.Length == 4) {
				string chargeInfo = values[3];
				if (chargeInfo[0] == '+' || chargeInfo[0] == '-')
				{
					int i = 1;
					while (i < chargeInfo.Length && char.IsDigit(chargeInfo[i]))
						i++;
					chargeInfo = chargeInfo.Substring(0, i);
					// Single charges should just display a sign.
					if (chargeInfo.Length == 2 && chargeInfo[1] == '1')
						chargeInfo = chargeInfo.Substring(0, 1);
					else
						chargeInfo = chargeInfo.Substring(1) + chargeInfo.Substring(0, 1);
					canvas.DrawString(chargeInfo, atomFont, atomColor, x + textOffsetX / 2, y + atomTextOffsetY * (float) 1.7);
				}
			}
		}

		private void renderLabel()
		{
			string[] values = currentToken.Split(MOL_RNDR_DELIM);
			if (values.Length != 4)
				throw new Exception ("Invalid label token '" + currentToken + "' - must consist of 4 values.");

			float x = int.Parse(values[0]);
			float y = int.Parse(values[1]);
			if ( x < clipRegion.X || x > clipRegion.X + clipRegion.Width || y < clipRegion.Y || y > clipRegion.Y + clipRegion.Height )
				return;

			string alignment_s = values[2];
			string theLabel    = values[3];

			theLabel = theLabel.Replace("%2C", ",");
			theLabel = theLabel.Replace("%3A", ":");
			theLabel = theLabel.Replace("%25", "%");

			if (alignment_s.Length != 1)
				throw new Exception ("Invalid label alignment '" + alignment_s + "' - must be one character.");
			char alignment = alignment_s[0];

			float labelWidth = canvas.MeasureString(theLabel, labelFont).Width;

			x = (float) (rect.X + x * coordToPixelFactor);
			y = (float) (rect.Y + y * coordToPixelFactor);

			if (alignment == MOL_RNDR_LABEL_ALIGN_CENTER)
				x -= labelWidth / 2;
			else if (alignment == MOL_RNDR_LABEL_ALIGN_LEFT)
				x = x + 0;
			else if (alignment == MOL_RNDR_LABEL_ALIGN_RIGHT)
				x -= labelWidth;
			else
				throw new Exception ("Invalid label alignment '" + alignment + "'.");

			canvas.DrawString(theLabel, labelFont, labelBrush, x, y + labelTextOffsetY);
		}

		private void processRenderInfo(string renderInfo, Rectangle the_rect, float zoomFactor, Graphics the_canvas, int whatToDo)
		{
			if (renderInfo.Length == 0)
				return;

			string[] tokens = renderInfo.Split(MOL_RNDR_TAG_START);

			if (whatToDo == DO_CALCBB) {
				readHeader(tokens);
				coordToPixelFactor = zoomFactor / COORD_TO_PIXEL_SCALING * (1000 / coordScaling);
				calcBoundingBox();
				return;
			}
			else {
				if (whatToDo != DO_RENDER)
					throw new Exception("Internal bug: Invalid request sent to ProcessRenderInfo (" + whatToDo.ToString() + ").");
			}

			try {
				rect   = the_rect;
				canvas = the_canvas;

				atomFont = new Font(defaultFont.Name, (float) (defaultFont.Size * zoomFactor));

				if (oneLetterResidues)
					residueFont = new Font(defaultFont.Name, atomFont.Size, FontStyle.Bold);
				else
					residueFont = new Font(defaultFont.Name, (float) (atomFont.Size * 0.8), FontStyle.Bold);

				labelFont = new Font(defaultFont.Name, (float) (atomFont.Size * 0.8));

				atomTextWidth  = the_canvas.MeasureString("W", atomFont).Width;
				atomTextHeight = the_canvas.MeasureString("W", atomFont).Height;

				labelTextHeight = the_canvas.MeasureString("W", labelFont).Height;
				labelTextOffsetY = - labelTextHeight / 2;
				
				readHeader(tokens);
				coordToPixelFactor = zoomFactor / COORD_TO_PIXEL_SCALING * (1000 / coordScaling);
				calcBoundingBox();
				// Calculate local clip region boundaries. Pre-filtering bond and atom
				// coordinates instead of relying on GDI clipping speeds up rendering of
				// very large molecules (fewer relatively expensive graphics subsystem calls).
				if (Math.Abs(coordToPixelFactor) > 0.000001)
					clipRegion = new Rectangle(
						(int) Math.Round( (-rect.X - atomTextWidth * 8) / coordToPixelFactor ),
						(int) Math.Round( (-rect.Y - atomTextHeight * 5) / coordToPixelFactor ),
						clipRegion.Left + (int) Math.Round( (rect.Right - rect.X + atomTextWidth * 16)  / coordToPixelFactor ),
						clipRegion.Top  + (int) Math.Round( (rect.Bottom - rect.Y + atomTextHeight * 10) / coordToPixelFactor )
					);
				else
					clipRegion = new Rectangle(0, 0, (int) Math.Round(xMax), (int) Math.Round(yMax));

				residueCircleRadius = (float) Math.Round(0.45 * coordScaling * coordToPixelFactor);
//(** DEBUG:				renderFrame();

				tokenIdx = 2;
				nextToken(tokens);

				while (tag == MOL_RNDR_TAG_BOND) {
					try {
						renderBond();
						nextToken(tokens);
					} catch (Exception e) {
						throw new Exception ("Error rendering bond: " + e.ToString());
					}
				}

				while (tag == MOL_RNDR_TAG_ATOM) {
					try {
						renderAtom();
						nextToken(tokens);
					} catch (Exception e) {
						throw new Exception ("Error rendering atom: " + e.ToString());
					}
				}

				while (tag == MOL_RNDR_TAG_LABEL) {
					try {
						renderLabel();
						nextToken(tokens);
					} catch (Exception e) {
						throw new Exception ("Error rendering label: " + e.ToString());
					}
				}

				if (tag != MOL_RNDR_TAG_INFO_END)
					throw new Exception ("Unsupported render info tag '" + tag + "'.");

			} catch (Exception e) {
				the_canvas.DrawString(e.ToString(), new Font(FONT_NAME, ERROR_MESSAGE_FONT_SIZE), new SolidBrush(Color.Red), rect.X + 10, rect.Y + 10);
			}
		}

		public Rectangle boundingBox(string renderInfo, Rectangle the_rect, float zoomFactor, Graphics the_canvas)
		{
			try {
				processRenderInfo(renderInfo, the_rect, zoomFactor, the_canvas, DO_CALCBB);
				return new Rectangle(0, 0, boxWidth, boxHeight);
			} catch {
				return new Rectangle(0, 0, 1, 1);
			}
		}

		public void render(string renderInfo, Rectangle the_rect, float zoomFactor, Graphics the_canvas)
		{
			processRenderInfo(renderInfo, the_rect, zoomFactor, the_canvas, DO_RENDER);
		}

		public bool oneLetterResidues
		{
			get
			{
				return _oneLetterResidues;
			}
			set
			{
				_oneLetterResidues = value;
			}
		}

	}
}
