/**
 * 
  Protein renderer that parses Biochemfusion sequence rendering info
  and produces a graphical output on a System.Drawing.Graphics device.

  Supports sequence rendering info V1 only.

  Biochemfusion recommends that you use a fixed-width font to display
  sequences.

(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This C# source code unit (ProteinRenderer.cs) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/

using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Biochemfusion
{
	class ProteinRenderer
	{
		const int SEQ_RNDR_BITFLAG_MODRES = 1; // Bit 0 = modified residue.
		const int SEQ_RNDR_BITFLAG_DFORM  = 2; // Bit 1 = residue is d-form.
		const string SEQ_RNDR_DELIM		= ",";			// Value delimiter.
		const string SEQ_RNDR_TAG_START	= ":";			// Tag start.
		const string SEQ_RNDR_TAG_BOUNDBOX	= "BB";		// Bounding box.
		const string SEQ_RNDR_TAG_CHAIN		= "CH";		// Chain start.
		const string SEQ_RNDR_TAG_DISULFIDE	= "SS";		// Disulfide bridge.
		const string SEQ_RNDR_TAG_CROSSLINK	= "XL";		// Lactam cycles and other crosslinks.
		const string SEQ_RNDR_TAG_INFO_BEGIN	= "BCF_SEQ_RNDR_INFO_V1";
		const string SEQ_RNDR_TAG_INFO_END		= "ED";	// End of Data.

		int cellSize;
		int lineSpacing;
		int lettersPerLine;
		int numRows;
		int residueCount;
		//
		int boxWidth;
		int boxHeight;
		//
		float XScale;
		float YScale;
		float XOffset;
		//
		Rectangle rect;
		Graphics canvas;
		Font font;

		const int MAX_NUM_DIGITS = 5;
		const string MAX_NUM_DIGITS_FORMATSTR = "{0,5}";

		const int NUM_CHAIN_BACKGND_COLORS = 2;
		/* "const" */ Brush [] CHAIN_BACKGND_COLORS = new Brush[NUM_CHAIN_BACKGND_COLORS];

		const int NUM_CHAIN_HILITE_COLORS = 2;
		/* "const" */ Brush [] CHAIN_HILITE_COLORS = new Brush[NUM_CHAIN_HILITE_COLORS];

		const int NUM_DISULFIDE_COLORS = 3;
		/* "const" */ Pen [] DISULFIDE_COLORS = new Pen[NUM_DISULFIDE_COLORS];

		const int NUM_CYCLE_COLORS = 3;
		/* "const" */ Pen [] CYCLE_COLORS = new Pen[NUM_CYCLE_COLORS];

		public ProteinRenderer()
		{
			// Background colors use direct RGB values. None of the System.Drawing.Color constants
			// match Delphi's clMoneyGreen and clSkyBlue well.
			CHAIN_BACKGND_COLORS[0] = new SolidBrush(Color.FromArgb(192, 220, 192));
			CHAIN_BACKGND_COLORS[1] = new SolidBrush(Color.FromArgb(166, 202, 240)); // RGB values for Delphi's clSkyBlue. None of the Color-constants match well.

			CHAIN_HILITE_COLORS[0] = new SolidBrush(Color.Lime);
			CHAIN_HILITE_COLORS[1] = new SolidBrush(Color.Aqua);

			DISULFIDE_COLORS[0] = new Pen(Color.FromArgb(255, 0, 0));
			DISULFIDE_COLORS[1] = new Pen(Color.FromArgb(255, 100, 0));
			DISULFIDE_COLORS[2] = new Pen(Color.FromArgb(255, 0, 100));

			CYCLE_COLORS[0] = new Pen(Color.FromArgb(0, 220, 0));
			CYCLE_COLORS[1] = new Pen(Color.FromArgb(150, 220, 0));
			CYCLE_COLORS[2] = new Pen(Color.FromArgb(0, 220, 150));
		}
		
		private void check(bool condition, string errMsg)
		{
			if (!condition)
				throw new Exception(errMsg);
		}

		private float XtoPixels(int X)
		{
			return X * XScale;
		}

		private float YtoPixels(int Y)
		{
			return Y * YScale;
		}

		private float XtoScreen(int X)
		{
			return XtoPixels(X) + XOffset + rect.X;
		}

		private float YtoScreen(int Y)
		{
			return YtoPixels(Y) + YtoPixels(lineSpacing) + rect.Y;
		}

		private void readHeader(string[] tokens)
		{
			// Valid render info at all ?
			// Note: tokens[0] is blank since render info starts with a delimiter.
			check(
				tokens.Length > 1 && tokens[1] == SEQ_RNDR_TAG_INFO_BEGIN,
				"Rendering info must start with header '" + SEQ_RNDR_TAG_INFO_BEGIN + "'."
			);
			check(
				tokens.Length > 2 && tokens[2].Substring(0, 2) == SEQ_RNDR_TAG_BOUNDBOX,
				"Rendering header must be followed by a bounding box '" + SEQ_RNDR_TAG_BOUNDBOX + "'."
			);
			check(
				tokens[tokens.Length - 1] == SEQ_RNDR_TAG_INFO_END,
				"Rendering info must end with an End-of-Data tag '" + SEQ_RNDR_TAG_INFO_END + "'."
			);

			// Read bounding box values.
			string[] values = tokens[2].Split(',');
			check(values.Length == 5, "Invalid bounding box - expected 5 values.");
			try {
			  cellSize       = int.Parse(values[0].Remove(0, 2));
			  lineSpacing    = int.Parse(values[1]);
			  lettersPerLine = int.Parse(values[2]);
			  numRows        = int.Parse(values[3]);
			  residueCount   = int.Parse(values[4]);
			} catch (Exception e) {
				throw new Exception ("Error reading bounding box: " + e.ToString());
			}
		}

		private void calcScalingAndBoundingBox()
		{
			XScale  = canvas.MeasureString("W", font).Width / cellSize * (1 + lineSpacing / cellSize);
			YScale  = canvas.MeasureString("W", font).Height / cellSize;
			XOffset = (int) Math.Round(canvas.MeasureString("W", font).Width * MAX_NUM_DIGITS, MidpointRounding.AwayFromZero);

			boxWidth  = (int) Math.Round(XtoPixels(cellSize * ((int)Math.Min(residueCount, lettersPerLine) + 1)) + XOffset, MidpointRounding.AwayFromZero);
			boxHeight = (int) Math.Round(YtoPixels((cellSize + lineSpacing) * numRows + lineSpacing), MidpointRounding.AwayFromZero);
		}

		private void renderFrame()
		{
			Pen pen = new Pen(new SolidBrush(Color.Black));
			canvas.DrawLine(pen, new Point(rect.X, rect.Y), new Point(rect.X + boxWidth, rect.Y));
			canvas.DrawLine(pen, new Point(rect.X + boxWidth, rect.Y), new Point(rect.X + boxWidth, rect.Y + boxHeight));
			canvas.DrawLine(pen, new Point(rect.X, rect.Y + boxHeight), new Point(rect.X + boxWidth, rect.Y + boxHeight));
			canvas.DrawLine(pen, new Point(rect.X, rect.Y + boxHeight), new Point(rect.X, rect.Y));
		}

		private void renderNumbers()
		{
			Brush brush = new SolidBrush(Color.Black);
			for (int i = 0; i < numRows; i++)
			{
				canvas.DrawString(
					String.Format(MAX_NUM_DIGITS_FORMATSTR, i * lettersPerLine + 1),
					font, brush,
					XtoScreen(-MAX_NUM_DIGITS * cellSize + cellSize / 2), YtoScreen(i * (cellSize + lineSpacing) + 1)
				);
			}
		}

		private float resPosX(int resIdx, int XOffset)
		{
			return XtoScreen( (resIdx % lettersPerLine) * cellSize + XOffset );
		}

		private float resPosY(int resIdx)
		{
			return YtoScreen( (resIdx / lettersPerLine) * (cellSize + lineSpacing));
		}

		private void renderTermModMarks(int colorIdx, int NtermFlag, int CtermFlag, int fromResidueIndex, int toResidueIndex)
		{
			float rectWidth = Math.Max(XtoPixels(cellSize) / 4, 2);

			if (NtermFlag != 0)
			{
				float X = resPosX(fromResidueIndex, 0);
				float Y = resPosY(fromResidueIndex);
				canvas.FillRectangle(CHAIN_HILITE_COLORS[colorIdx % NUM_CHAIN_HILITE_COLORS], X, Y, rectWidth, YtoPixels(cellSize));
			}
			if (CtermFlag != 0)
			{
				float X = resPosX(toResidueIndex, cellSize);
				float Y = resPosY(toResidueIndex);
				canvas.FillRectangle(CHAIN_HILITE_COLORS[colorIdx % NUM_CHAIN_HILITE_COLORS], X - rectWidth, Y, rectWidth, YtoPixels(cellSize));
			}
		}
		
		private void renderChainBackground(int colorIdx, int NtermFlag, int CtermFlag, int fromResidueIndex, int toResidueIndex)
		{
			Brush chainColor = CHAIN_BACKGND_COLORS[colorIdx % NUM_CHAIN_BACKGND_COLORS];

			float rectHeight   = YtoPixels(cellSize) + 1;
			int savedFromIdx   = fromResidueIndex;
			int savedToIdx     = toResidueIndex;

			while (fromResidueIndex <= toResidueIndex)
			{
				int colFrom = fromResidueIndex % lettersPerLine;
				int colTo   = colFrom + (toResidueIndex - fromResidueIndex);
				// Wrap onto next line ?
				if (colTo >= lettersPerLine)
					colTo = lettersPerLine - 1;

				float rectXFrom = XtoScreen(colFrom * cellSize);
				float rectXTo   = XtoScreen(colTo * cellSize + cellSize);
				float rectY     = YtoScreen( (fromResidueIndex / lettersPerLine) * (cellSize + lineSpacing) );
				canvas.FillRectangle(chainColor, rectXFrom, rectY, rectXTo - rectXFrom, rectHeight);

				fromResidueIndex += colTo - colFrom + 1;
			}

			renderTermModMarks(colorIdx, NtermFlag, CtermFlag, savedFromIdx, savedToIdx);
		}

		private void renderChain(string token, ref int residueIndex)
		{
			string[] values = token.Split(',');
			check(values.Length == 3, "Invalid chain token - expected 3 values.");

			int chainNo         = int.Parse(values[0].Remove(0, 2));
			string termModFlags = values[1];
			string chain        = values[2];

			check(chain.Length > 0, "Empty chain.");
			check(termModFlags.Length == 2, "Invalid terminal modification flags - must be two characters.");
			int NtermFlag = int.Parse(termModFlags.Substring(0, 1));
			int CtermFlag = int.Parse(termModFlags.Substring(1, 1));
			check(NtermFlag == 0 || NtermFlag == 1, "Invalid N-terminal modification flag.");
			check(CtermFlag == 0 || CtermFlag == 1, "Invalid C-terminal modification flag.");

			char[] modFlags = new string('0', chain.Length).ToCharArray();
			for (int i = 0; i < chain.Length; i++)
			{
				char c = chain[i];
				if (c == '1' || c == '2' || c == '3')
				{
					modFlags[i] = chain[i];
					chain = chain.Remove(i, 1);
				}
			}

			if (chainNo != 0)
				renderChainBackground((chainNo - 1) % NUM_CHAIN_BACKGND_COLORS, NtermFlag, CtermFlag, residueIndex, residueIndex + chain.Length - 1);

			// Render amino acid codes.
			// Make sure that letter is placed in the middle of the cell.
			float textXOffset = canvas.MeasureString("W", font).Width / 2;

			int virtX = (residueIndex % lettersPerLine) * cellSize + cellSize / 2;
			int virtY = (residueIndex / lettersPerLine) * (cellSize + lineSpacing);

			Brush textBrushBlack = new SolidBrush(Color.Black);
			Brush textBrushRed = new SolidBrush(Color.Red);
			for (int i = 0; i < chain.Length; i++)
			{
				int modFlag = modFlags[i] - '0';
				if ((modFlag & SEQ_RNDR_BITFLAG_DFORM) != 0)
				{
					Pen pen = new Pen(Color.White);
					pen.Width = Math.Max(XtoPixels(cellSize) / 6, 3);

					canvas.DrawLine(pen,
						XtoScreen(virtX - cellSize / 2), YtoScreen(virtY + cellSize) - 2,
						XtoScreen(virtX - cellSize / 2 + cellSize), YtoScreen(virtY) + 2
					);
				}

				Brush textBrush = textBrushBlack;
				if ((modFlag & SEQ_RNDR_BITFLAG_MODRES) != 0)
					textBrush = textBrushRed;

				canvas.DrawString(
					chain.Substring(i, 1), font, textBrush,
					XtoScreen(virtX) - textXOffset, YtoScreen(virtY + 1)
				);

				virtX += cellSize;
				if (virtX > cellSize * lettersPerLine)
				{
					virtX = cellSize / 2;
					virtY += cellSize + lineSpacing;
				}
			}

			residueIndex += chain.Length;
		}

		private void renderLine(string token, int values_to_skip, Pen pen)
		{
			string[] values = token.Split(',');
			pen.Width = Math.Max(XtoPixels(cellSize) / 4, 2);

			// Start at index 'values_to_skip' (discarding 'from' and 'to' and other non-line-drawing information).
			PointF [] points = new PointF [(values.Length - values_to_skip) / 2];
			int valueIdx = values_to_skip;
			for (int pointIdx = 0; pointIdx < points.Length; pointIdx++)
			{
				points[pointIdx].X = XtoScreen( int.Parse(values[valueIdx]) );
				points[pointIdx].Y = YtoScreen( int.Parse(values[valueIdx + 1]) );
				valueIdx += 2;
			}
			canvas.DrawLines(pen, points);
		}

		public void render(string renderInfo, Rectangle the_rect, Graphics the_canvas, Font the_font)
		{
			if (renderInfo.Length == 0)
			{
				the_canvas.DrawString(
					"<No rendering info to display>", the_font, new SolidBrush(Color.Gray),
					rect.X + 10, rect.Y + 10
				);
				return;
			}

			try {
				rect = the_rect;
				canvas = the_canvas;
				font = the_font;

				string[] tokens = renderInfo.Split(':');
				readHeader(tokens);
				calcScalingAndBoundingBox();

				// renderFrame();	// Show only when debugging/testing.
				renderNumbers();

				int residueIndex   = 0;
				int disulfideCount = 0;
				int cycleCount     = 0;

				for (int curTokenIdx = 3; curTokenIdx < tokens.Length - 1; curTokenIdx++)
				{
					string currentToken = tokens[curTokenIdx];
					string tag = currentToken.Substring(0, 2);
					currentToken.Remove(0, 2);

					if (tag == SEQ_RNDR_TAG_CHAIN)
					{
						try {
							renderChain(currentToken, ref residueIndex);
						} catch (Exception e) {
							throw new Exception("Error rendering chain: " + e.ToString());
						}
					}
					else
					if (tag == SEQ_RNDR_TAG_DISULFIDE)
					{
						try {
						  renderLine(currentToken, 2, DISULFIDE_COLORS[disulfideCount % NUM_DISULFIDE_COLORS]);
						  disulfideCount++;
						} catch (Exception e) {
							throw new Exception("Error rendering disulfide bridge: " + e.ToString());
						}
					}
					else
					if (tag == SEQ_RNDR_TAG_CROSSLINK)
					{
						try {
							// Check validity of terminal flags, but ignore them for now.
							string[] values = currentToken.Split(',');
							string cycleTerminalFlags = values[0].Remove(0, 2);
							check(cycleTerminalFlags.Length == 2, "Invalid cyclization terminal flags.");
							check(cycleTerminalFlags[0] == '0' || cycleTerminalFlags[0] == '1', "Invalid cyclization terminal flag.");
							check(cycleTerminalFlags[1] == '0' || cycleTerminalFlags[1] == '1', "Invalid cyclization terminal flag.");

							renderLine(currentToken, 3, CYCLE_COLORS[cycleCount % NUM_CYCLE_COLORS]);
							cycleCount++;
						} catch (Exception e) {
							throw new Exception("Error rendering cycle: " + e.ToString());
						}
					}
					else
						throw new Exception("Unsupported render info tag '" + tag + "'.");
				}
			} catch (Exception e) {
				canvas.DrawString(e.ToString(), font, new SolidBrush(Color.Red), rect.X + 10, rect.Y + 10);
			}
		}


		
		public Rectangle boundingBox(string renderInfo, Graphics the_canvas, Font the_font)
		{
			Rectangle result = new Rectangle(0, 0, 1, 1);
			if (renderInfo.Length == 0)
				return result;

			canvas = the_canvas;
			font = the_font;

			string[] tokens = renderInfo.Split(':');
			readHeader(tokens);
			calcScalingAndBoundingBox();

			result.Width  = boxWidth;
			result.Height = boxHeight;

			return result;
		}

	}

}
