/**
  Flash molecule renderer that parses molecule rendering info V2 from the
  Biochemfusion Proteax library and produces a graphical output.
  
(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This Flash source code unit (MoleculeRenderer.as) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/
package { 

import flash.display.DisplayObjectContainer;
import flash.display.Graphics;
import flash.display.Shape;
import flash.display.Sprite;
import flash.text.*;

import flash.events.*;
import flash.display.StageAlign;
import flash.display.StageScaleMode;

public class MoleculeRenderer { 

	// Info passed to main public function.
	private var canvasContainer:DisplayObjectContainer;
	// Two containers: One for the background and one for the molecule drawing; both are contained in canvasContainer.
	private var backgroundContainer:Sprite;
	private var moleculeContainer:Sprite;
	
	private var aCanvas:Graphics;
	private var renderInfo:String;
	// Resulting bounding box size in pixels.
	private var xBBoxSize:int;
	private var yBBoxSize:int;

	// Default font and its metrics for all text output via textOut().
	private var fontFormat:TextFormat;
	private var fontMetrics:TextLineMetrics;
	private var residueFontFormat:TextFormat;
	private var residueFontMetrics:TextLineMetrics;
	// For parsing render info input.
	private var fromPos:int;
	private var toPos:int;

	private function textOut(x:int, y:int, textCaption:String, textColor:int = 0x000000):void {
		var txt:TextField = new TextField();

		fontFormat.color = textColor;
		txt.defaultTextFormat = fontFormat;
		txt.text = textCaption;
		txt.autoSize = TextFieldAutoSize.LEFT;

		var metrics:TextLineMetrics = txt.getLineMetrics(0);
		txt.x = x - metrics.width / 2 - metrics.x;
		txt.y = y - (metrics.ascent + metrics.descent) / 2 - 2; // 2 pixel Y-gutter.

		// Just doing a moleculeContainer.graphics.drawRect() to blank out the background freaks out on the Android phone.
		// But this approach works fine even though it seems somewhat wasteful with all these text fields.
		var bckgnd:TextField = new TextField();
		bckgnd.background = true;
		bckgnd.backgroundColor = 0xFFFFFF;
		bckgnd.x = txt.x + metrics.x - 1;
		bckgnd.y = txt.y + 1;
		bckgnd.width = metrics.width + 2;
		bckgnd.height = metrics.height;
		moleculeContainer.addChild(bckgnd);
		
		moleculeContainer.addChild(txt);
	}

	private function textOutNoBkgnd(x:int, y:int, textCaption:String, aTextFormat:TextFormat, textColor:int = 0x000000, textAlign:String = TextFieldAutoSize.CENTER):void {
		var txt:TextField = new TextField();
		fontFormat.color = textColor;
		txt.defaultTextFormat = aTextFormat;
		txt.text = textCaption;
		txt.autoSize = TextFieldAutoSize.LEFT;

		var metrics:TextLineMetrics = txt.getLineMetrics(0);
		txt.y = y - (metrics.ascent + metrics.descent) / 2 - 2; // 2 pixel Y-gutter.

		if (textAlign == TextFieldAutoSize.LEFT)
			txt.x = x - metrics.x;
		else if (textAlign == TextFieldAutoSize.RIGHT)
			txt.x = x - metrics.width - metrics.x;
		else if (textAlign == TextFieldAutoSize.CENTER)
			txt.x = x - metrics.width / 2 - metrics.x;
		else
			throw Error('Unsupported text alignment "' + textAlign + '".');

		moleculeContainer.addChild(txt);
	}

	// Handle resizing of stage area - scale molecule to fit window when window gets smaller than molecule drawing.
	private function stageResizeHandler(e: Event): void {
		var scale:Number = Math.min(canvasContainer.stage.stageWidth / xBBoxSize, canvasContainer.stage.stageHeight / yBBoxSize);
		scale = Math.min(scale, 1.0);
		moleculeContainer.scaleX = scale;		
		moleculeContainer.scaleY = scale;
		moleculeContainer.x = Math.max((canvasContainer.stage.stageWidth - xBBoxSize * scale) / 2, 0);
		moleculeContainer.y = Math.max((canvasContainer.stage.stageHeight - yBBoxSize * scale) / 2, 0);
		
		backgroundContainer.scaleX = canvasContainer.stage.stageWidth;
		backgroundContainer.scaleY = canvasContainer.stage.stageHeight;
	}

	private function check(condition:Boolean, errMsg:String):void {
		if (!condition)
			throw Error(errMsg);
	}

	public function renderMolecule(the_zoomFactor:Number, the_renderInfo:String, the_target:DisplayObjectContainer, the_canvas:Graphics):void {
		canvasContainer = the_target;
		aCanvas         = the_canvas;
		renderInfo      = the_renderInfo;
		
		canvasContainer.stage.scaleMode = StageScaleMode.NO_SCALE;
		canvasContainer.stage.align = StageAlign.TOP_LEFT;

		fontFormat = new TextFormat();
		fontFormat.font  = "Arial";
		fontFormat.color = 0x000000;
		fontFormat.size  = 10;
		
		residueFontFormat = new TextFormat();
		residueFontFormat.font = "Arial";
		residueFontFormat.color = 0x000000;
		residueFontFormat.size  = 9;

		var tempTextField:TextField = new TextField();
		tempTextField.defaultTextFormat = fontFormat;
		tempTextField.text = "W";
		fontMetrics = tempTextField.getLineMetrics(0);
	
		tempTextField = new TextField();
		tempTextField.defaultTextFormat = residueFontFormat;
		tempTextField.text = "W";
		residueFontMetrics = tempTextField.getLineMetrics(0);
	
		// Add a nicely colored background rectangle.
		backgroundContainer = new Sprite();
		backgroundContainer.x = 0;
		backgroundContainer.y = 0;
		backgroundContainer.graphics.beginFill(0xFFFFFF);
		backgroundContainer.graphics.lineStyle(1, 0xFFFFFF);
		backgroundContainer.graphics.drawRect(0, 0, 1, 1);
		backgroundContainer.graphics.endFill();
		canvasContainer.addChild(backgroundContainer);
		
		// Create a place to render molecule.
		moleculeContainer = new Sprite();
		canvasContainer.addChild(moleculeContainer);
		aCanvas = moleculeContainer.graphics;

		if (the_renderInfo.length == 0) {
			textOut(10, 10, "<No rendering info to display>");
			return;
		}
		
		try {

		var textHeight:int = fontMetrics.height;

		var render_tokens:Array = the_renderInfo.split(':');
		if (render_tokens.length < 2)
			throw Error('No rendering info or malformed rendering info.');
		if (render_tokens[1] != 'BCF_MOL_RNDR_INFO_V2')
			throw Error('Rendering info does not start with an "BCF_MOL_RNDR_INFO_V2" header tag.');
		if (render_tokens[render_tokens.length - 1] != 'ED')
			throw Error('Rendering info does not end with an "ED" end tag.');

		// Read bounding box and calculate coordinate scaling factor.
		var bbParams:Array = render_tokens[2].slice(2).split(',');
		xBBoxSize = bbParams[1];
		yBBoxSize = bbParams[2];
		var coordScalingFactor:Number = bbParams[0];
		var scaleFactor:Number = the_zoomFactor * 1.4 / 60 * (1000 / coordScalingFactor);
		xBBoxSize = xBBoxSize * scaleFactor;
		yBBoxSize = yBBoxSize * scaleFactor;

/**
		// Render bounding box. Use for debug purposes only.
		aCanvas.lineStyle(1, 0x000000);
		aCanvas.beginFill(0xFFFFFF);
		aCanvas.drawRect(0, 0, xBBoxSize, yBBoxSize);
		aCanvas.endFill();
**/

		var values:Array;
		
		var x:Number;
		var y:Number;
		var textWidth:int;
		
		var residueCircleRadius:Number = 0.45 * coordScalingFactor * scaleFactor;

		aCanvas.lineStyle(1, 0x000000);
		// Skip first three elements - a blank (split() perceives a blank token at start since the rendering info starts with a delimiter),
		// the header and the bounding box.
		var tokenIdx:int = 3;

		// Render bonds.
		while (render_tokens[tokenIdx].substring(0, 2) == 'BD') {
			values = render_tokens[tokenIdx].split(',');
			var x1:Number = values[0].slice(2);
			var y1:Number = values[1];
			var x2:Number = values[2];
			var y2:Number = values[3];
			x1 = x1 * scaleFactor;
			y1 = y1 * scaleFactor;
			x2 = x2 * scaleFactor;
			y2 = y2 * scaleFactor;
			var cardinality:String = values[4];
			var extraInfo:String   = cardinality.charAt(1);
			cardinality            = cardinality.slice(0, 1);
			
			var orthoX:Number;
			var orthoY:Number;
			if (cardinality == '1') {
				if (extraInfo == '.') {
					aCanvas.moveTo(x1, y1);
					aCanvas.lineTo(x2, y2);
				}
				else {
					orthoX = -(y2 - y1) * 0.1;
					orthoY =  (x2 - x1) * 0.1;
					aCanvas.moveTo(x1, y1);
					if (extraInfo == 'u') {
						// 'Up'-bonds are black.
						aCanvas.lineStyle(1, 0x000000);
						aCanvas.beginFill(0x000000);
					}
					if (extraInfo == 'd') {
						// 'Down'-bonds are light-grey. //(**!! TODO: Use pattern to get proper down-bond style.
						aCanvas.lineStyle(1, 0xAAAAAA);
						aCanvas.beginFill(0xAAAAAA);
					}
					aCanvas.lineTo(x2 - orthoX, y2 - orthoY);
					aCanvas.lineTo(x2 + orthoX, y2 + orthoY);
					aCanvas.endFill();
					aCanvas.lineStyle(1, 0x000000);
				}
			}
			else
			if (cardinality == '2') {
				orthoX = -(y2 - y1) * 0.1;
				orthoY =  (x2 - x1) * 0.1;
				if (extraInfo != 's') {
					orthoX = orthoX * 1.6;
					orthoY = orthoY * 1.6;
					var x1a:Number = x1 + orthoY;
					var y1a:Number = y1 - orthoX;
					var x2a:Number = x2 - orthoY;
					var y2a:Number = y2 + orthoX;
				}
				if (extraInfo == 's') {
					aCanvas.moveTo(x1 - orthoX, y1 - orthoY);
					aCanvas.lineTo(x2 - orthoX, y2 - orthoY);
					aCanvas.moveTo(x1 + orthoX, y1 + orthoY);
					aCanvas.lineTo(x2 + orthoX, y2 + orthoY);
				}
				if (extraInfo == 'l') {
					// 'Left' dbl-bond.
					aCanvas.moveTo(x1, y1);
					aCanvas.lineTo(x2, y2);
					aCanvas.moveTo(x1a - orthoX, y1a - orthoY);
					aCanvas.lineTo(x2a - orthoX, y2a - orthoY);
				}
				if (extraInfo == 'r') {
					// 'Right' dbl-bond.
					aCanvas.moveTo(x1, y1);
					aCanvas.lineTo(x2, y2);
					aCanvas.moveTo(x1a + orthoX, y1a + orthoY);
					aCanvas.lineTo(x2a + orthoX, y2a + orthoY);
				}
			}
			else
			if (cardinality == '3') {
				orthoX = -(y2 - y1) * 0.16;
				orthoY =  (x2 - x1) * 0.16;
				aCanvas.moveTo(x1 - orthoX, y1 - orthoY);
				aCanvas.lineTo(x2 - orthoX, y2 - orthoY);
				aCanvas.moveTo(x1, y1);
				aCanvas.lineTo(x2, y2);
				aCanvas.moveTo(x1 + orthoX, y1 + orthoY);
				aCanvas.lineTo(x2 + orthoX, y2 + orthoY);
			}
			else
			if (cardinality == 'a') {
				aCanvas.lineStyle(2, 0xFF0000);
				aCanvas.moveTo(x1, y1);
				aCanvas.lineTo(x2, y2);
				aCanvas.lineStyle(1, 0x000000);

				if (extraInfo == '1')
					textOutNoBkgnd(x2, y2, '1* ', fontFormat, 0xFF0000, TextFieldAutoSize.RIGHT);
				else
				if (extraInfo == '2')
					textOutNoBkgnd(x2 + 2, y2, '*2', fontFormat, 0xFF0000, TextFieldAutoSize.LEFT);
				else
					throw Error('Unsupported APO number (' + extraInfo + ').');
			}
			else
				throw Error('Unsupported bond cardinality "' + cardinality + '".');
			
			tokenIdx++;
		}
		
		// Render atoms.
		while (render_tokens[tokenIdx].substring(0, 2) == 'AT') {
			values = render_tokens[tokenIdx].split(',');
			x = values[0].slice(2);
			y = values[1];
			var atomSymbol:String = values[2];
			extraInfo = '';
			if (values.length > 3)
				extraInfo = values[3];

			x = x * scaleFactor;
			y = y * scaleFactor;

			// All 3-letter atoms will be assumed to be pseudo atoms / residue atoms. Not fool-proof but usually good enough.
			var isPseudoAtom:Boolean = atomSymbol.length == 3;

			if (isPseudoAtom)
				textWidth = residueFontMetrics.width * (atomSymbol.length);
			else
				textWidth = fontMetrics.width * (atomSymbol.length);

			if (isPseudoAtom) {
				aCanvas.lineStyle(1, 0x000000);
				// Default chain number is 1.
				var chainNumber:int = 1;
				if (extraInfo != '')
					chainNumber = int(extraInfo);
				if (chainNumber % 2 == 0)
					aCanvas.beginFill(0xa6ccff); // clSkyBlue
				else
					aCanvas.beginFill(0xc0e0c0); // clMoneyGreen
				aCanvas.drawCircle(x, y, residueCircleRadius);
				aCanvas.endFill();
				textOutNoBkgnd(x, y, atomSymbol, residueFontFormat);
			}
			else {
				if (atomSymbol == 'R1' || atomSymbol == 'R#')
					textOut(x, y, atomSymbol, 0xFF0000);
				else
					textOut(x, y, atomSymbol);
			}

			// Handle charge info if it is there.
			if (!isPseudoAtom && extraInfo.length > 0) {
				if (extraInfo.charAt(0) == '+' || extraInfo.charAt(0) == '-') {
					var chargeInfo:String = extraInfo;
					var i :int = 1;
					while (chargeInfo.charAt(i).match(/\d/) && i < chargeInfo.length)
						i++;
					chargeInfo = chargeInfo.substring(0, i);
					if (chargeInfo.length == 2 && chargeInfo.charAt(1) == '1')
						chargeInfo = chargeInfo.charAt(0);
					else
						chargeInfo = chargeInfo.substring(1) + chargeInfo.charAt(0);
					textOutNoBkgnd(x + fontMetrics.width / 2, y - textHeight / 2, chargeInfo, fontFormat, 0x000000, TextFieldAutoSize.LEFT);
				}
			}

			tokenIdx++;
		}
		
		// Render labels.
		while (render_tokens[tokenIdx].substring(0, 2) == 'LB') {
			values = render_tokens[tokenIdx].split(',');
			x = values[0].slice(2);
			y = values[1];
			var alignment:String = values[2];
			var labelText:String = values[3];
			labelText = labelText.replace(/%3A/g, ':');
			labelText = labelText.replace(/%2C/g, ',');
			labelText = labelText.replace(/%25/g, '%');

			x = x * scaleFactor;
			y = y * scaleFactor;

			if (alignment == 'l')
				textOutNoBkgnd(x, y, labelText, fontFormat, 0x000000, TextFieldAutoSize.LEFT);
			else if (alignment == 'r')
				textOutNoBkgnd(x, y, labelText, fontFormat, 0x000000, TextFieldAutoSize.RIGHT);
			else if (alignment == 'c')
				textOutNoBkgnd(x, y, labelText, fontFormat, 0x000000, TextFieldAutoSize.CENTER);
			else
				throw Error('Unsupported text alignment tag "' + alignment + '".');
			
			tokenIdx++;
		}

		if (render_tokens[tokenIdx] != 'ED')
			throw Error('Unsupported render info tag "' + render_tokens[tokenIdx].substring(0, 2) + '".');
		
		} catch (e:Error) {
			textOut(10, 10, String(e));
		}

		canvasContainer.stage.addEventListener(Event.RESIZE, stageResizeHandler);
		stageResizeHandler(null);
	}		

}

}
