/**

  2D molecule renderer that parses molecule rendering info V2 produced by
  Proteax and generates corresponding graphical output on an AWT Graphics
  abstraction.

(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This Java source code unit (MoleculeRenderer.java) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion ApS (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/

package com.biochemfusion.proteax;

import java.io.*;

import java.awt.*;
import java.awt.font.*;

public class MoleculeRenderer {

	private static Color SKY_BLUE = new Color(166, 204, 255);
	private static Color MONEY_GREEN = new Color(192, 224, 192);

	private void setLineWidth(Graphics g, int lineWidth)
	{
		Graphics2D g2d = (Graphics2D) g;
		g2d.setStroke(new BasicStroke(lineWidth));
	}
	
	private void drawLine(Graphics g, double x1, double y1, double x2, double y2)
	{
		g.drawLine((int) Math.round(x1), (int) Math.round(y1), (int) Math.round(x2), (int) Math.round(y2));
	}
	
	private void drawString(Graphics g, String s, double x, double y)
	{
		g.drawString(s, (int) Math.round(x), (int) Math.round(y));
	}

	/**
	  Note: This implementation assumes that you have already scaled the font according to 
	  the zoomFactor. An Arial 10pt font is suitable for a zoomFactor of 1.0.
	**/
	public void render(Graphics g, double zoomFactor, String renderInfo)
	{
		try {
		
		String [] renderTokens = renderInfo.split(":");
		// There must be at least 4 elements: A blank, a header, a bounding box, and an end tag.
		if (renderTokens.length < 4)
			throw new Exception("No rendering info or malformed rendering info.");
		if (!renderTokens[1].equals("BCF_MOL_RNDR_INFO_V2"))
			throw new Exception("Rendering info does not start with an \"BCF_MOL_RNDR_INFO_V2\" header tag.");
		if (!renderTokens[renderTokens.length - 1].equals("ED"))
			throw new Exception("Rendering info does not end with an \"ED\" end tag.");

		// Read bounding box rectangle and coordinate scaling factor.
		String [] values = renderTokens[2].substring(2).split(",");
		double xMax = Double.parseDouble(values[1]);
		double yMax = Double.parseDouble(values[2]);
		double coordScalingFactor = Double.parseDouble(values[0]);
		double scaleFactor = zoomFactor * 1.4 / 60 * (1000 / coordScalingFactor);
		
		double residueCircleRadius = 0.45 * coordScalingFactor * scaleFactor;
		int residueCircleDiameter  = (int) Math.round(2 * residueCircleRadius);
		
		g.setColor(Color.black);

		// Skip the first three elements: a blank (split() perceives a blank token at start since the rendering info starts with a delimiter),
		// the header and the bounding box.
		int tokenIdx = 3;

		// Render bonds.
		while (renderTokens[tokenIdx].substring(0, 2).equals("BD")) {
			values = renderTokens[tokenIdx].split(",");
			double x1 = Double.parseDouble(values[0].substring(2));
			double y1 = Double.parseDouble(values[1]);
			double x2 = Double.parseDouble(values[2]);
			double y2 = Double.parseDouble(values[3]);
			char bondType    = values[4].charAt(1);
			char cardinality = values[4].charAt(0);

			x1 = x1 * scaleFactor;
			y1 = y1 * scaleFactor;
			x2 = x2 * scaleFactor;
			y2 = y2 * scaleFactor;
			
			if (cardinality == '1') {
				if (bondType == '.') {
					drawLine(g, x1, y1, x2, y2);
				}
				else {
					double orthoX = -(y2 - y1) * 0.1;
					double orthoY =  (x2 - x1) * 0.1;

					if (bondType == 'u')
						// 'Up'-bonds are black.
						g.setColor(Color.black);
					if (bondType == 'd')
						// 'Down'-bonds are light-grey. //(**!! TODO: Use pattern to get proper down-bond style.
						g.setColor(Color.lightGray);

					Polygon p = new Polygon();
					p.addPoint((int) Math.round(x1), (int) Math.round(y1));
					p.addPoint((int) Math.round(x2 - orthoX), (int) Math.round(y2 - orthoY));
					p.addPoint((int) Math.round(x2 + orthoX), (int) Math.round(y2 + orthoY));
					g.drawPolygon(p);
					g.fillPolygon(p);
					
					g.setColor(Color.black);
				}
			}
			else
			if (cardinality == '2') {
				double orthoX = -(y2 - y1) * 0.1;
				double orthoY =  (x2 - x1) * 0.1;
				double x1a = x1, y1a = y1, x2a = x2, y2a = y2;
				if (bondType != 's') {
					orthoX = orthoX * 1.6;
					orthoY = orthoY * 1.6;
					x1a = x1 + orthoY;
					y1a = y1 - orthoX;
					x2a = x2 - orthoY;
					y2a = y2 + orthoX;
				}
				if (bondType == 's') {
					drawLine(g, x1 - orthoX, y1 - orthoY, x2 - orthoX, y2 - orthoY);
					drawLine(g, x1 + orthoX, y1 + orthoY, x2 + orthoX, y2 + orthoY);
				}
				if (bondType == 'l') {
					// 'Left' dbl-bond.
					drawLine(g, x1, y1, x2, y2);
					drawLine(g, x1a - orthoX, y1a - orthoY, x2a - orthoX, y2a - orthoY);
				}
				if (bondType == 'r') {
					// 'Right' dbl-bond.
					drawLine(g, x1, y1, x2, y2);
					drawLine(g, x1a + orthoX, y1a + orthoY, x2a + orthoX, y2a + orthoY);
				}
			}
			else
			if (cardinality == '3') {
				double orthoX = -(y2 - y1) * 0.16;
				double orthoY =  (x2 - x1) * 0.16;
				drawLine(g, x1 - orthoX, y1 - orthoY, x2 - orthoX, y2 - orthoY);
				drawLine(g, x1, y1, x2, y2);
				drawLine(g, x1 + orthoX, y1 + orthoY, x2 + orthoX, y2 + orthoY);
			}
			else
			if (cardinality == 'a') {
				g.setColor(Color.red);
				setLineWidth(g, 2);
				drawLine(g, x1, y1, x2, y2);

				FontMetrics fm = g.getFontMetrics();
				LineMetrics lm = fm.getLineMetrics("1*", g);
				
				double textWidth = fm.stringWidth("1*");
				y2 += (lm.getAscent() - lm.getDescent()) / 2;
				
				if (bondType == '1')
					drawString(g, "1*", x2 - textWidth * 1.4, y2);
				else
				if (bondType == '2')
					drawString(g, "*2", x2 + textWidth * 0.4, y2);
				else
					throw new Exception("Invalid APO number (" + bondType + ").");

				setLineWidth(g, 1);
				g.setColor(Color.black);
			}
			else
				throw new Exception("Unsupported bond cardinality \"" + cardinality + "\".");

			tokenIdx++;
		}

		// Render atoms.
		while (renderTokens[tokenIdx].substring(0, 2).equals("AT")) {
			values = renderTokens[tokenIdx].split(",");
			double x = Double.parseDouble(values[0].substring(2));
			double y = Double.parseDouble(values[1]);
			String atomSymbol = values[2];
			String extraInfo = "";
			if (values.length > 3)
				extraInfo = values[3];

			x = x * scaleFactor;
			y = y * scaleFactor;

			FontMetrics fm = g.getFontMetrics();
			double textWidth = fm.stringWidth(atomSymbol);
			LineMetrics lm = fm.getLineMetrics(atomSymbol, g);
			double textHeight = lm.getAscent() + lm.getDescent();

			// All 3-letter atoms will be assumed to be pseudo atoms / residue atoms. Not fool-proof but usually good enough.
			// TODO: Should be looked up in a table; comes naturally if we add support for one/three-letter display of residue atoms as in the Pascal renderer.
			boolean isPseudoAtom = atomSymbol.length() == 3;
			if (isPseudoAtom) {
				int chainNumber = 1;
				if (extraInfo.length() > 0)
					chainNumber = Integer.parseInt(extraInfo);
				if (chainNumber % 2 == 0)
					g.setColor(SKY_BLUE);
				else
					g.setColor(MONEY_GREEN);

				int x1 = (int) Math.round(x - residueCircleRadius);
				int y1 = (int) Math.round(y - residueCircleRadius);
				g.fillOval(x1, y1, residueCircleDiameter, residueCircleDiameter);
				g.setColor(Color.black);
				g.drawOval(x1, y1, residueCircleDiameter, residueCircleDiameter);
			}
			else {
				g.setColor(Color.white);
				
				Polygon p = new Polygon();
				double x2 = x - textWidth / 2;
				double y2 = y - textHeight / 2;
				p.addPoint((int) Math.round(x2), (int) Math.round(y2));
				p.addPoint((int) Math.round(x2 + textWidth), (int) Math.round(y2));
				p.addPoint((int) Math.round(x2 + textWidth), (int) Math.round(y2 + textHeight));
				p.addPoint((int) Math.round(x2), (int) Math.round(y2 + textHeight));
				g.drawPolygon(p);
				g.fillPolygon(p);

				g.setColor(Color.black);
			}

			// Centre text both horizontally and vertically.
			x -= textWidth / 2;
			y += (lm.getAscent() - lm.getDescent()) / 2;

			if (atomSymbol.equals("R#") || atomSymbol.equals("R1"))
				g.setColor(Color.red);

			drawString(g, atomSymbol, x, y);
			g.setColor(Color.black);

			// Handle charge info if it is there.
			if (!isPseudoAtom && extraInfo.length() != 0) {
				if (extraInfo.charAt(0) == '+' || extraInfo.charAt(0) == '-') {
					String chargeInfo = extraInfo;
					int i = 1;
					while (i < chargeInfo.length() && Character.isDigit( chargeInfo.charAt(i) ))
						i++;
					chargeInfo = chargeInfo.substring(0, i);
					// Single charges should just display a sign.
					if (chargeInfo.length() == 2 && chargeInfo.charAt(1) == '1')
						chargeInfo = chargeInfo.substring(0, 1);
					else
						chargeInfo = chargeInfo.substring(1) + chargeInfo.substring(0, 1);
					drawString(g, chargeInfo, x + textWidth, y - textHeight / 2);
				}
			}
			
			tokenIdx++;
		}

		// Render labels.
		while (renderTokens[tokenIdx].substring(0, 2).equals("LB")) {
			values = renderTokens[tokenIdx].split(",");
			double x = Double.parseDouble(values[0].substring(2));
			double y = Double.parseDouble(values[1]);
			String alignment = values[2];
			String labelText = values[3];

			labelText = labelText.replaceAll("%3A", ":");
			labelText = labelText.replaceAll("%2C", ",");
			labelText = labelText.replaceAll("%25", "%");

			x = x * scaleFactor;
			y = y * scaleFactor;

			FontMetrics fm = g.getFontMetrics();
			double textWidth = fm.stringWidth(labelText);
			// Centre text vertically.
			LineMetrics lm = fm.getLineMetrics(labelText, g);
			y += (lm.getAscent() - lm.getDescent()) / 2;

			if (alignment.equals("l"))
				x = x;
			else if (alignment.equals("r"))
				x = x - textWidth;
			else if (alignment.equals("c"))
				x = x - textWidth / 2;
			else
				throw new Exception("Unsupported text alignment tag \"" + alignment + "\".");
			
			drawString(g, labelText, x, y);

			tokenIdx++;
		}

		if (!renderTokens[tokenIdx].equals("ED"))
			throw new Exception("Unsupported render info tag \"" + renderTokens[tokenIdx].substring(0, 2) + "\".");

		} catch (Exception e) {
			StringWriter sw = new StringWriter();
			e.printStackTrace(new PrintWriter(sw));
			String trace = sw.toString();
			if (trace.length() > 250)
				trace = trace.substring(0, 250) + "...";

			g.setColor(Color.red);
			g.drawString(trace, 10, 20);
		}
	}
	
	public Dimension boundingBox(Graphics g, double zoomFactor, String renderInfo) throws Exception
	{
		Dimension result = new Dimension(1, 1);
		
		String [] renderTokens = renderInfo.split(":");
		// There must be at least 4 elements: A blank, a header, a bounding box, and an end tag.
		if (renderTokens.length < 4)
			throw new Exception("No rendering info or malformed rendering info.");
		if (!renderTokens[1].equals("BCF_MOL_RNDR_INFO_V2"))
			throw new Exception("Rendering info does not start with an \"BCF_MOL_RNDR_INFO_V2\" header tag.");
		if (!renderTokens[renderTokens.length - 1].equals("ED"))
			throw new Exception("Rendering info does not end with an \"ED\" end tag.");

		// Read bounding box rectangle and coordinate scaling factor.
		String [] values = renderTokens[2].substring(2).split(",");
		double xMax = Double.parseDouble(values[1]);
		double yMax = Double.parseDouble(values[2]);
		double coordScalingFactor = Double.parseDouble(values[0]);
		double scaleFactor = zoomFactor * 1.4 / 60 * (1000 / coordScalingFactor);

		result.width  = (int) Math.round(xMax * scaleFactor);
		result.height = (int) Math.round(yMax * scaleFactor);

		return result;
	}

};
