/**

  Sequence renderer that parses sequence rendering info V1 produced by
  Proteax and generates corresponding graphical output on an AWT Graphics
  abstraction.

(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This Java source code unit (ProteinRenderer.java) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion ApS (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/

package com.biochemfusion.proteax;

import java.io.*;

import java.awt.*;
import java.awt.font.*;

public class ProteinRenderer {

	private final static int SEQ_RNDR_BITFLAG_MODRES = 1; // Bit 0 = modified residue.
	private final static int SEQ_RNDR_BITFLAG_DFORM  = 2; // Bit 1 = residue is d-form.
	private final static String SEQ_RNDR_DELIM		= ",";			// Value delimiter.
	private final static String SEQ_RNDR_TAG_START	= ":";			// Tag start.
	private final static String SEQ_RNDR_TAG_BOUNDBOX	= "BB";		// Bounding box.
	private final static String SEQ_RNDR_TAG_CHAIN		= "CH";		// Chain start.
	private final static String SEQ_RNDR_TAG_DISULFIDE	= "SS";		// Disulfide bridge.
	private final static String SEQ_RNDR_TAG_CROSSLINK	= "XL";		// Lactam cycles and other crosslinks.
	private final static String SEQ_RNDR_TAG_INFO_BEGIN	= "BCF_SEQ_RNDR_INFO_V1";
	private final static String SEQ_RNDR_TAG_INFO_END		= "ED";	// End of Data.

	private int cellSize;
	private int lineSpacing;
	private int lettersPerLine;
	private int numRows;
	private int residueCount;
	// Bounding box dimensions.
	private int boxWidth;
	private int boxHeight;

	private double XScale;
	private double YScale;
	private double XOffset;
	// Font metrics in pixels.
	private double charWidth;
	private double charHeight;
	private double charHalfSpacing;	// Half-width of (forced) inter-character spacing.
	private double charAscent;
	private double charDescent;
	//
	private Graphics canvas;
	
	private final static int MAX_NUM_DIGITS = 5;
	private final static String MAX_NUM_DIGITS_FORMATSTR = "%" + MAX_NUM_DIGITS + "d";

	private final static int NUM_CHAIN_BACKGND_COLORS = 2;
	private final static Color [] CHAIN_BACKGND_COLORS = { new Color(192, 224, 192), new Color(166, 204, 255) }; // clMoneyGreen, clSkyBlue

	private final static int NUM_CHAIN_HILITE_COLORS = 2;
	private final static Color [] CHAIN_HILITE_COLORS = { new Color(0, 255, 0), new Color(0, 255, 255) }; // clLime, clAqua

	private final static int NUM_DISULFIDE_COLORS = 3;
	private final static Color [] DISULFIDE_COLORS = { new Color(255, 0, 0), new Color(255, 100, 0), new Color(255, 0, 100) };

	private final static int NUM_CYCLE_COLORS = 3;
	private final static Color [] CYCLE_COLORS = { new Color(0, 220, 0), new Color(150, 220, 0), new Color(0, 220, 150) };

	private void check(boolean condition, String errMsg) throws Exception
	{
		if (!condition)
			throw new Exception (errMsg);
	}

	private int XtoPixels(int X)
	{
		return (int) Math.round(X * XScale);
	}

	private int YtoPixels(int Y)
	{
		return (int) Math.round(Y * YScale);
	}

	private int XtoScreen(int X)
	{
		return (int) Math.round(X * XScale + XOffset);
	}

	private int YtoScreen(int Y)
	{
		return (int) Math.round((Y + lineSpacing) * YScale);
	}

	private void setLineWidth(int lineWidth)
	{
		Graphics2D g2d = (Graphics2D) canvas;
		g2d.setStroke(new BasicStroke(lineWidth));
	}
	
	private void drawLine(Graphics g, double x1, double y1, double x2, double y2)
	{
		g.drawLine((int) Math.round(x1), (int) Math.round(y1), (int) Math.round(x2), (int) Math.round(y2));
	}
	
	private void fillRectangle(Graphics g, Color color, double x, double y, double width, double height)
	{
		g.setColor(color);
		
		Polygon p = new Polygon();
		p.addPoint((int) Math.round(x), (int) Math.round(y));
		p.addPoint((int) Math.round(x + width), (int) Math.round(y));
		p.addPoint((int) Math.round(x + width), (int) Math.round(y + height));
		p.addPoint((int) Math.round(x), (int) Math.round(y + height));

		g.fillPolygon(p);
		g.drawPolygon(p);
		
		g.setColor(Color.black);
	}
	
	private void drawString(Graphics g, String s, double x, double y)
	{
		g.drawString(s, (int) Math.round(x), (int) Math.round(y));
	}

	private void readHeader(String[] tokens) throws Exception
	{
		// Valid render info at all ?
		// Note: tokens[0] is blank since render info starts with a delimiter.
		check(
			tokens.length > 1 && tokens[1].equals(SEQ_RNDR_TAG_INFO_BEGIN),
			"Rendering info must start with header '" + SEQ_RNDR_TAG_INFO_BEGIN + "'."
		);
		check(
			tokens.length > 2 && tokens[2].substring(0, 2).equals(SEQ_RNDR_TAG_BOUNDBOX),
			"Rendering header must be followed by a bounding box '" + SEQ_RNDR_TAG_BOUNDBOX + "'."
		);
		check(
			tokens[tokens.length - 1].equals(SEQ_RNDR_TAG_INFO_END),
			"Rendering info must end with an End-of-Data tag '" + SEQ_RNDR_TAG_INFO_END + "'."
		);

		// Read bounding box values.
		String[] values = tokens[2].split(",");
		check(values.length == 5, "Invalid bounding box - expected 5 values.");
		try {
		  cellSize       = Integer.parseInt(values[0].substring(2));
		  lineSpacing    = Integer.parseInt(values[1]);
		  lettersPerLine = Integer.parseInt(values[2]);
		  numRows        = Integer.parseInt(values[3]);
		  residueCount   = Integer.parseInt(values[4]);
		} catch (Exception e) {
			throw new Exception ("Error reading bounding box: " + e.toString());
		}
	}

	private void calcScalingAndBoundingBox()
	{	
		FontMetrics fm = canvas.getFontMetrics();
		LineMetrics lm = fm.getLineMetrics("W", canvas);
		//(**!! Don't know how to get a text width reported that includes character spacing ? Got to do the spacing myself.
		double spacing = 1.7;
		charWidth         = fm.stringWidth("W") * spacing;
		charHeight        = lm.getAscent() + lm.getDescent();
		
		charHalfSpacing   = fm.stringWidth("W") * (spacing - 1) / 2;
		charAscent        = lm.getAscent();
		charDescent       = lm.getDescent();
		
		XScale  = charWidth / cellSize * (1 + lineSpacing / cellSize);
		YScale  = charHeight / cellSize;
		XOffset = (int) Math.round(charWidth * MAX_NUM_DIGITS);

		boxWidth  = (int) Math.round(XtoPixels(cellSize * ((int)Math.min(residueCount, lettersPerLine) + 1)) + XOffset);
		boxHeight = (int) Math.round(YtoPixels((cellSize + lineSpacing) * numRows + lineSpacing));
	}

	private void renderFrame()
	{
		canvas.setColor(Color.black);
		drawLine(canvas, 0, 0, boxWidth, 0);
		drawLine(canvas, boxWidth, 0, boxWidth, boxHeight);
		drawLine(canvas, 0, boxHeight, boxWidth, boxHeight);
		drawLine(canvas, 0, boxHeight, 0, 0);
	}

	private void renderNumbers()
	{
		canvas.setColor(Color.black);
		for (int i = 0; i < numRows; i++)
		{
			drawString(canvas,
				String.format(MAX_NUM_DIGITS_FORMATSTR, i * lettersPerLine + 1),
				XtoScreen(-MAX_NUM_DIGITS * cellSize + cellSize / 2), YtoScreen(i * (cellSize + lineSpacing) + 1) + charAscent
			);
		}
	}

	private double resPosX(int resIdx, int XOffset)
	{
		return XtoScreen( (resIdx % lettersPerLine) * cellSize + XOffset );
	}

	private double resPosY(int resIdx)
	{
		return YtoScreen( (resIdx / lettersPerLine) * (cellSize + lineSpacing));
	}

	private void renderTermModMarks(int colorIdx, int NtermFlag, int CtermFlag, int fromResidueIndex, int toResidueIndex)
	{
		double rectWidth = Math.max(XtoPixels(cellSize) / 4, 2);

		if (NtermFlag != 0)
		{
			double X = resPosX(fromResidueIndex, 0);
			double Y = resPosY(fromResidueIndex);
			fillRectangle(canvas, CHAIN_HILITE_COLORS[colorIdx % NUM_CHAIN_HILITE_COLORS], X, Y, rectWidth, YtoPixels(cellSize));
		}
		if (CtermFlag != 0)
		{
			double X = resPosX(toResidueIndex, cellSize);
			double Y = resPosY(toResidueIndex);
			fillRectangle(canvas, CHAIN_HILITE_COLORS[colorIdx % NUM_CHAIN_HILITE_COLORS], X - rectWidth, Y, rectWidth, YtoPixels(cellSize));
		}
	}

	private void renderChainBackground(int colorIdx, int NtermFlag, int CtermFlag, int fromResidueIndex, int toResidueIndex)
	{
		Color chainColor = CHAIN_BACKGND_COLORS[colorIdx % NUM_CHAIN_BACKGND_COLORS];

		double rectHeight  = YtoPixels(cellSize) + 1;
		int savedFromIdx   = fromResidueIndex;
		int savedToIdx     = toResidueIndex;

		while (fromResidueIndex <= toResidueIndex)
		{
			int colFrom = fromResidueIndex % lettersPerLine;
			int colTo   = colFrom + (toResidueIndex - fromResidueIndex);
			// Wrap onto next line ?
			if (colTo >= lettersPerLine)
				colTo = lettersPerLine - 1;

			double rectXFrom = XtoScreen(colFrom * cellSize);
			double rectXTo   = XtoScreen(colTo * cellSize + cellSize);
			double rectY     = YtoScreen( (fromResidueIndex / lettersPerLine) * (cellSize + lineSpacing) );
			fillRectangle(canvas, chainColor, rectXFrom, rectY, rectXTo - rectXFrom, rectHeight);

			fromResidueIndex += colTo - colFrom + 1;
		}

		renderTermModMarks(colorIdx, NtermFlag, CtermFlag, savedFromIdx, savedToIdx);
	}

	// In Java, renderChain() returns the new value of 'residueIndex' since Java
	// does not support pass-by-reference.
	private int renderChain(String token, int residueIndex) throws Exception
	{
		String[] values = token.split(",");
		check(values.length == 3, "Invalid chain token - expected 3 values.");

		int chainNo         = Integer.parseInt(values[0]);
		String termModFlags = values[1];
		String chain        = values[2];

		check(chain.length() > 0, "Empty chain.");
		check(termModFlags.length() == 2, "Invalid terminal modification flags - must be two characters.");
		int NtermFlag = Integer.parseInt(termModFlags.substring(0, 1));
		int CtermFlag = Integer.parseInt(termModFlags.substring(1, 2));
		check(NtermFlag == 0 || NtermFlag == 1, "Invalid N-terminal modification flag.");
		check(CtermFlag == 0 || CtermFlag == 1, "Invalid C-terminal modification flag.");

		int[] modFlags = new int[chain.length()];
		for (int i = 0; i < chain.length(); i++)
		{
			char c = chain.charAt(i);
			if (c == '1' || c == '2' || c == '3')
			{
				modFlags[i] = c - '0';
				chain = chain.substring(0, i) + chain.substring(i + 1, chain.length()); // Remove .charAt(i)
			}
		}

		if (chainNo != 0)
			renderChainBackground((chainNo - 1) % NUM_CHAIN_BACKGND_COLORS, NtermFlag, CtermFlag, residueIndex, residueIndex + chain.length() - 1);

		// Render amino acid codes.
		// Make sure that letter is placed in the middle of the cell.
		double textXOffset = charWidth / 2 - charHalfSpacing;

		int virtX = (residueIndex % lettersPerLine) * cellSize + cellSize / 2;
		int virtY = (residueIndex / lettersPerLine) * (cellSize + lineSpacing);

		for (int i = 0; i < chain.length(); i++)
		{
			int modFlag = modFlags[i] - '0';
			if ((modFlag & SEQ_RNDR_BITFLAG_DFORM) != 0)
			{
				canvas.setColor(Color.white);
				setLineWidth( Math.max(XtoPixels(cellSize) / 6, 3) );

				drawLine(canvas,
					XtoScreen(virtX - cellSize / 2), YtoScreen(virtY + cellSize) - 2,
					XtoScreen(virtX - cellSize / 2 + cellSize), YtoScreen(virtY) + 2
				);
			}

			canvas.setColor(Color.black);
			if ((modFlag & SEQ_RNDR_BITFLAG_MODRES) != 0)
				canvas.setColor(Color.red);

			drawString(canvas,
				chain.substring(i, i + 1), (int) Math.round(XtoScreen(virtX) - textXOffset), YtoScreen(virtY + 1) + charAscent
			);

			virtX += cellSize;
			if (virtX > cellSize * lettersPerLine)
			{
				virtX = cellSize / 2;
				virtY += cellSize + lineSpacing;
			}
		}

		residueIndex += chain.length();
		return residueIndex;
	}

	private void renderLine(String token, int values_to_skip, Color lineColor)
	{
		String[] values = token.split(",");

		setLineWidth(Math.max(XtoPixels(cellSize) / 4, 2));
		canvas.setColor(lineColor);

		// Start at index 'values_to_skip' (discarding 'from' and 'to' and other non-line-drawing information).
		int valueIdx = values_to_skip;
		int x1, x2, y1, y2;
		x1 = XtoScreen( Integer.parseInt(values[valueIdx]) );
		y1 = YtoScreen( Integer.parseInt(values[valueIdx + 1]) );

		valueIdx += 2;
		while (valueIdx < values.length) {
			x2 = XtoScreen( Integer.parseInt(values[valueIdx]) );
			y2 = YtoScreen( Integer.parseInt(values[valueIdx + 1]) );
			canvas.drawLine(x1, y1, x2, y2);
			x1 = x2;
			y1 = y2;
			valueIdx += 2;
		}
		
		canvas.setColor(Color.black);
	}

	
	public void render(Graphics g, String renderInfo)
	{
		if (renderInfo.length() == 0) {
			g.drawString("<No rendering info to display>", 10, 10);
			return;
		}

		try {
		
		canvas = g;

		String [] tokens = renderInfo.split(":");
		readHeader(tokens);
		calcScalingAndBoundingBox();

		// renderFrame();	// Show only when debugging/testing.
		renderNumbers();

		int residueIndex   = 0;
		int disulfideCount = 0;
		int cycleCount     = 0;

		for (int curTokenIdx = 3; curTokenIdx < tokens.length - 1; curTokenIdx++)
		{
			String currentToken = tokens[curTokenIdx];
			String tag   = currentToken.substring(0, 2);
			currentToken = currentToken.substring(2);

			if (tag.equals(SEQ_RNDR_TAG_CHAIN))
			{
				try {
					residueIndex = renderChain(currentToken, residueIndex);
				} catch (Exception e) {
					throw new Exception("Error rendering chain: " + e.toString());
				}
			}
			else
			if (tag.equals(SEQ_RNDR_TAG_DISULFIDE))
			{
				try {
				  renderLine(currentToken, 2, DISULFIDE_COLORS[disulfideCount % NUM_DISULFIDE_COLORS]);
				  disulfideCount++;
				} catch (Exception e) {
					throw new Exception("Error rendering disulfide bridge: " + e.toString());
				}
			}
			else
			if (tag.equals(SEQ_RNDR_TAG_CROSSLINK))
			{
				try {
					// Check validity of terminal flags, but ignore them for now.
					String[] values = currentToken.split(",");
					String cycleTerminalFlags = values[0].substring(0, 2);
					check(cycleTerminalFlags.length() == 2, "Invalid cyclization terminal flags.");
					check(cycleTerminalFlags.charAt(0) == '0' || cycleTerminalFlags.charAt(0) == '1', "Invalid cyclization terminal flag.");
					check(cycleTerminalFlags.charAt(1) == '0' || cycleTerminalFlags.charAt(1) == '1', "Invalid cyclization terminal flag.");

					renderLine(currentToken, 3, CYCLE_COLORS[cycleCount % NUM_CYCLE_COLORS]);
					cycleCount++;
				} catch (Exception e) {
					throw new Exception("Error rendering cycle: " + e.toString());
				}
			}
			else
				throw new Exception("Unsupported render info tag '" + tag + "'.");
		}
		
		} catch (Exception e) {
			StringWriter sw = new StringWriter();
			e.printStackTrace(new PrintWriter(sw));
			String trace = sw.toString();
			if (trace.length() > 250)
				trace = trace.substring(0, 250) + "...";

			g.setColor(Color.red);
			g.drawString(trace, 10, 20);
		}
	}


	public Rectangle boundingBox(Graphics g, String renderInfo) throws Exception
	{
		Rectangle result = new Rectangle(0, 0, 1, 1);
		if (renderInfo.length() == 0)
			return result;

		canvas = g;

		String[] tokens = renderInfo.split(":");
		readHeader(tokens);
		calcScalingAndBoundingBox();

		result.width  = boxWidth;
		result.height = boxHeight;

		return result;
	}


};
