/**
JavaScript code for parsing and rendering Biochemfusion molecule
rendering info text V2.

(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This JavaScript source code unit (bcf_mol_renderer.js) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/

function render_molecule(canvas_id, zoomFactor, render_info, scale_canvas) {
	var render_tokens = render_info.split(':');
	// There must be at least 4 elements: A blank, a header, a bounding box, and an end tag.
	if (render_tokens.length < 4)
		throw 'No rendering info or malformed rendering info.';
	if (render_tokens[1] != 'BCF_MOL_RNDR_INFO_V2')
		throw 'Rendering info does not start with an "BCF_MOL_RNDR_INFO_V2" header tag.';
	if (render_tokens[render_tokens.length - 1] != 'ED')
		throw 'Rendering info does not end with an "ED" end tag.';

	// Read bounding box rectangle and coordinate scaling factor.
	var values = render_tokens[2].slice(2).split(',')
	var xMax = values[1];
	var yMax = values[2];
	var coordScalingFactor = values[0];
	var scaleFactor = zoomFactor * 1.4 / 60 * (1000 / coordScalingFactor);

	var canvas = document.getElementById(canvas_id);
	if (scale_canvas) {
		canvas.width  = xMax * scaleFactor;
		canvas.height = yMax * scaleFactor;
	}
	if (!canvas.getContext)
		return;
	
	var ctx = canvas.getContext('2d');
	
	ctx.strokeStyle = '#000000';
	ctx.beginPath();
	
	ctx.textAlign    = 'center';
	ctx.textBaseline = 'middle';
	ctx.fillStyle    = '#ff2200';

	var textHeight = Math.round(10 * zoomFactor);
	ctx.font = textHeight + "px sans-serif";
	
	var residueCircleRadius = 0.45 * coordScalingFactor * scaleFactor;

	// Skip first three elements - a blank (split() perceives a blank token at start since the rendering info starts with a delimiter),
	// the header and the bounding box.
	var tokenIdx = 3;

	// Render bonds.
	while (render_tokens[tokenIdx].substring(0, 2) == 'BD') {
		var values = render_tokens[tokenIdx].split(',');
		var x1 = values[0].slice(2);
		var y1 = values[1];
		var x2 = values[2];
		var y2 = values[3];
		var cardinality = values[4];

		x1 = x1 * scaleFactor;
		y1 = y1 * scaleFactor;
		x2 = x2 * scaleFactor;
		y2 = y2 * scaleFactor;
		var extraInfo   = cardinality.charAt(1);
		cardinality     = cardinality.slice(0, 1);
		
		if (cardinality == '1') {
			if (extraInfo == '.') {
				ctx.moveTo(x1, y1);
				ctx.lineTo(x2, y2);
			}
			else {
				var orthoX = -(y2 - y1) * 0.1;
				var orthoY =  (x2 - x1) * 0.1;
				if (extraInfo == 'u')
					// 'Up'-bonds are black.
					ctx.fillStyle = '#000000';
				if (extraInfo == 'd')
					// 'Down'-bonds are light-grey. //(**!! TODO: Use pattern to get proper down-bond style.
					ctx.fillStyle = '#cccccc';
				// Must stroke what has previously been drawn as we begin a new (fill) path.
				ctx.stroke();
				ctx.closePath();
				ctx.beginPath();
				ctx.moveTo(x1, y1);
				ctx.lineTo(x2 - orthoX, y2 - orthoY);
				ctx.lineTo(x2 + orthoX, y2 + orthoY);
				ctx.closePath();
				ctx.fill();
				ctx.beginPath();
			}
		}
		else
		if (cardinality == '2') {
			var orthoX = -(y2 - y1) * 0.1;
			var orthoY =  (x2 - x1) * 0.1;
			if (extraInfo != 's') {
				orthoX = orthoX * 1.6;
				orthoY = orthoY * 1.6;
				var x1a = x1 + orthoY;
				var y1a = y1 - orthoX;
				var x2a = x2 - orthoY;
				var y2a = y2 + orthoX;
			}
			if (extraInfo == 's') {
				ctx.moveTo(x1 - orthoX, y1 - orthoY);
				ctx.lineTo(x2 - orthoX, y2 - orthoY);
				ctx.moveTo(x1 + orthoX, y1 + orthoY);
				ctx.lineTo(x2 + orthoX, y2 + orthoY);
			}
			if (extraInfo == 'l') {
				// 'Left' dbl-bond.
				ctx.moveTo(x1, y1);
				ctx.lineTo(x2, y2);
				ctx.moveTo(x1a - orthoX, y1a - orthoY);
				ctx.lineTo(x2a - orthoX, y2a - orthoY);
			}
			if (extraInfo == 'r') {
				// 'Right' dbl-bond.
				ctx.moveTo(x1, y1);
				ctx.lineTo(x2, y2);
				ctx.moveTo(x1a + orthoX, y1a + orthoY);
				ctx.lineTo(x2a + orthoX, y2a + orthoY);
			}
		}
		else
		if (cardinality == '3') {
			var orthoX = -(y2 - y1) * 0.16;
			var orthoY =  (x2 - x1) * 0.16;
			ctx.moveTo(x1 - orthoX, y1 - orthoY);
			ctx.lineTo(x2 - orthoX, y2 - orthoY);
			ctx.moveTo(x1, y1);
			ctx.lineTo(x2, y2);
			ctx.moveTo(x1 + orthoX, y1 + orthoY);
			ctx.lineTo(x2 + orthoX, y2 + orthoY);
		}
		else
		if (cardinality == 'a') {
			var savedTextAlign   = ctx.textAlign;
			var savedStrokeStyle = ctx.strokeStyle;
			var savedFillStyle   = ctx.fillStyle;
			var savedLineWidth   = ctx.lineWidth;
			
			ctx.stroke();
			ctx.closePath();
			ctx.beginPath();
			
			ctx.strokeStyle = "#ff0000"
			ctx.fillStyle = "#ff0000"
			ctx.lineWidth = 2 * ctx.lineWidth
			ctx.moveTo(x1, y1);
			ctx.lineTo(x2, y2);
			ctx.stroke();
			
			ctx.closePath();
			ctx.fill();
			ctx.beginPath();

			if (extraInfo == '1') {
				ctx.textAlign = 'right';
				ctx.fillText('1* ', x2, y2);
			}
			else
			if (extraInfo == '2') {
				ctx.textAlign = 'left';
				ctx.fillText(' *2', x2, y2);
			}
			else
				throw 'Invalid APO number (' + extraInfo + ').'

			ctx.textAlign   = savedTextAlign;
			ctx.strokeStyle = savedStrokeStyle;
			ctx.fillStyle   = savedFillStyle;
			ctx.lineWidth   = savedLineWidth;
		}
		else
			throw 'Unsupported bond cardinality "' + cardinality + '".';

		tokenIdx++;
	}
	// Finish bond drawing if a path is still open.
	ctx.stroke();
	ctx.closePath();

	// Render atoms.
	while (render_tokens[tokenIdx].substring(0, 2) == 'AT') {
		var values = render_tokens[tokenIdx].split(',');
		var x = values[0].slice(2);
		var y = values[1];
		var atomSymbol = values[2];
		var extraInfo  = '';
		if (values.length > 3)
			extraInfo = values[3];

		x = x * scaleFactor;
		y = y * scaleFactor;
		var textWidth = ctx.measureText(atomSymbol).width;

		// All 3-letter atoms will be assumed to be pseudo atoms / residue atoms. Not fool-proof but usually good enough.
		var isPseudoAtom = atomSymbol.length == 3;
		if (isPseudoAtom) {
			ctx.strokeStyle = "#000000";
			// Default chain number = 1.
			if (extraInfo === '')
				extraInfo = 1;
			if (extraInfo % 2 == 0)
				ctx.fillStyle = "#a6ccff"; // clSkyBlue
			else
				ctx.fillStyle = "#c0e0c0"; // clMoneyGreen
			ctx.beginPath();
			ctx.arc(x, y, residueCircleRadius, 0, Math.PI*2, true);
			ctx.closePath();
			ctx.stroke();
			ctx.fill();
		}
		else {
			ctx.fillStyle = '#ffffff';
			ctx.fillRect(x - textWidth / 2, y - textHeight / 2, textWidth, textHeight);
		}

		if (atomSymbol == 'R#' || atomSymbol == 'R1')
			ctx.fillStyle = '#ff0000';
		else
			ctx.fillStyle = '#000000';
		ctx.fillText(atomSymbol, x, y);
		// Handle charge info if it is there.
		if (!isPseudoAtom && extraInfo != '') {
			if (extraInfo.charAt(0) == '+' || extraInfo.charAt(0) == '-') {
				var chargeInfo = extraInfo;
				var i = 1;
				while (chargeInfo.charAt(i).match(/\d/) && i < chargeInfo.length)
					i++;
				chargeInfo = chargeInfo.slice(0, i);

				if (chargeInfo.length == 2 && chargeInfo.charAt(1) == '1')
					chargeInfo = chargeInfo.charAt(0);
				else
					chargeInfo = chargeInfo.slice(1) + chargeInfo.charAt(0);

				var savedTextAlign = ctx.textAlign;
				ctx.textAlign = 'left';
				ctx.fillText(chargeInfo, x + textWidth / 2, y - textHeight / 2);
				ctx.textAlign = savedTextAlign;
			}
		}
		
		tokenIdx++;
	}

	// Render labels.
	ctx.fillStyle = '#000000';
	while (render_tokens[tokenIdx].substring(0, 2) == 'LB') {
		var values = render_tokens[tokenIdx].split(',');
		var x = values[0].slice(2);
		var y = values[1];
		var alignment = values[2];
		var labelText = values[3];
		labelText = labelText.replace(/%3A/g, ':');
		labelText = labelText.replace(/%2C/g, ',');
		labelText = labelText.replace(/%25/g, '%');

		x = x * scaleFactor;
		y = y * scaleFactor;
		var textWidth = ctx.measureText(labelText).width;

		if (alignment == 'l')
			ctx.textAlign = 'left';
		else if (alignment == 'r')
			ctx.textAlign = 'right';
		else if (alignment == 'c')
			ctx.textAlign = 'center';
		else
			throw 'Unsupported text alignment tag "' + alignment + '".'
		
		ctx.fillText(labelText, x, y);

		tokenIdx++;
	}

	if (render_tokens[tokenIdx] != 'ED')
			throw 'Unsupported render info tag "' + render_tokens[tokenIdx].substring(0, 2) + '".'

}
