/**
JavaScript code for parsing and rendering Biochemfusion sequence
rendering info text V1.

Originally written by jan@biochemfusion.com, 2010.

(C) 2010 Jan Holst Jensen, jan@biochemfusion.com.

This JavaScript source code unit (bcf_seq_renderer.js) is released under a BSD-style
license:

* Copyright (C) 2010, Biochemfusion (http://www.biochemfusion.com)
* All rights reserved.
*
* Redistribution and use for any purpose in source and binary forms, with or
* without modification, are permitted, subject to the following restrictions:
*
* 1. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software. If you use this software
*    in a product, an acknowledgment in the product documentation would be
*    appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
*    misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution. 
*
* THIS SOFTWARE IS PROVIDED BY Biochemfusion ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Biochemfusion BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/

function render_sequence(canvas_id, render_info, font_height, scale_canvas) {
	var MAX_NUM_DIGITS = 5;
	var PAD_STRING = "     ";
	
	var NUM_DISULFIDE_COLORS = 3;
	var DISULFIDE_COLORS = ["#ff0000", "#ff6400", "#ff0064"];

	var NUM_CYCLE_COLORS = 3;
	var CYCLE_COLORS = ["#00dc00", "#96dc00", "#00dc96"];

	var NUM_CHAIN_BACKGND_COLORS = 2;
	var CHAIN_BACKGND_COLORS = ["#c0dcc0", "#a6caf0"];

	var NUM_CHAIN_HILITE_COLORS = 2;
	var CHAIN_HILITE_COLORS = ["#00ff00", "#00ffff"];

	function xToPixels(x)
	{
		return x * xScale;
	}
	
	function yToPixels(y)
	{
		return y * yScale;
	}
	
	function xToScreen(x)
	{
		return xToPixels(x) + xOffset;
	}
	
	function yToScreen(y)
	{
		return yToPixels(y + lineSpacing);
	}
	
	// Draw text at virtual coordinates (x_v, y_v).
	function drawText(x_v, y_v, txt)
	{
		ctx.fillText(
			txt,
			xToScreen(x_v + cellSize * 0.25),
			yToScreen(y_v + cellSize * 0.25) + charAscent
		);
	}
	
	// Debug function for testing how mapping between virtual coordinates and screen coordinates works for both text and lines.
	function renderTestGraphics(do_cells, do_text)
	{
		ctx.strokeStyle = "#ff0000";
		ctx.fillStyle = '#ff0000';
		if (do_cells) {
			// Draw character cells.
			for (var row = 0; row < numRows; row++) {
				for (var col = 0; col < lettersPerLine; col++) {
					ctx.strokeRect(
						xToScreen(col * cellSize),
						yToScreen(row * (cellSize + lineSpacing)),
						xToPixels(cellSize),
						yToPixels(cellSize) - 1
					);
				}
			}
		}
		ctx.stroke();
		ctx.fillStyle = "#000000";
		if (do_text) {
			// Draw test letters.
			var test = "A234567890B234567890C234567890C234567890E2C4567890F234567890G234567890H234567890I234567890J234CC789CK2345678C0";
			var stringIdx = 0;
			for (var row = 0; row < numRows; row++) {
				for (var col = 0; col < lettersPerLine; col++) {
					drawText(col * cellSize, row * (cellSize + lineSpacing), test.charAt(stringIdx));
					stringIdx++;
					if (stringIdx >= test.length)
						break;
				}
			}
		}
		ctx.stroke();
	}

	function renderNumbers()
	{
		for (var i = 0; i < numRows; i++) {
			// There is no equivalent of Format() or sprintf() in JavaScript so we'll just pad the strings ourselves.
			var outStr = (i * lettersPerLine + 1) + '';
			outStr = PAD_STRING.slice(0, MAX_NUM_DIGITS - outStr.length) + outStr;
			drawText((-MAX_NUM_DIGITS) * cellSize, i * (cellSize + lineSpacing), outStr);
		}
	}
	
	function resPosX(resIdx, xOffset)
	{
		return xToScreen( (resIdx % lettersPerLine) * cellSize + xOffset );
	}

	function resPosY(resIdx)
	{
		return yToScreen(Math.floor(resIdx / lettersPerLine) * (cellSize + lineSpacing));
	}

	function renderTermModMarks(colorIdx, NtermFlag, CtermFlag, fromResidueIndex, toResidueIndex)
	{
		ctx.fillStyle = CHAIN_HILITE_COLORS[colorIdx % NUM_CHAIN_HILITE_COLORS];
		var rectWidth = Math.max(xToPixels(cellSize) / 4, 2);

		if (NtermFlag != 0) {
			var x = resPosX(fromResidueIndex, 0);
			var y = resPosY(fromResidueIndex);
			ctx.fillRect(x, y, rectWidth, yToPixels(cellSize));
		}
		if (CtermFlag != 0) {
			var x = resPosX(toResidueIndex, cellSize);
			var y = resPosY(toResidueIndex);
			ctx.fillRect(x - rectWidth, y, rectWidth, yToPixels(cellSize));
		}
	}

	function renderChainBackground(chainNo, NtermFlag, CtermFlag, fromResidueIndex, toResidueIndex)
	{
		ctx.fillStyle   = CHAIN_BACKGND_COLORS[(chainNo - 1) % NUM_CHAIN_BACKGND_COLORS];

		var rectHeight   = yToPixels(cellSize);
		var savedFromIdx = fromResidueIndex;
		var savedToIdx   = toResidueIndex;

		while (fromResidueIndex <= toResidueIndex) {
			var colFrom = fromResidueIndex % lettersPerLine;
			var colTo   = colFrom + (toResidueIndex - fromResidueIndex);
			// Wrap onto next line ?
			if (colTo >= lettersPerLine)
				colTo = lettersPerLine - 1;

			var rectXFrom = xToScreen(colFrom * cellSize);
			var rectXTo   = xToScreen(colTo * cellSize + cellSize);
			var rectY     = yToScreen(Math.floor(fromResidueIndex / lettersPerLine) * (cellSize + lineSpacing));

			ctx.fillRect(rectXFrom, rectY, rectXTo - rectXFrom, rectHeight);
		
			fromResidueIndex = fromResidueIndex + colTo - colFrom + 1;
		}
		
		renderTermModMarks(chainNo - 1, NtermFlag, CtermFlag, savedFromIdx, savedToIdx);
	}
	
	function renderChain()
	{
		var values = currentToken.split(",");
		if (values.length !== 3) throw "Invalid chain token - expected 3 values.";

		var chainNo      = parseInt(values[0]);
		var termModFlags = values[1];
		var chain        = values[2];

		if (chain.length === 0) throw "Empty chain.";
		if (termModFlags.length !== 2) throw "Invalid terminal modification flags - must be two characters.";
		var NtermFlag = parseInt(termModFlags.slice(0, 1));
		var CtermFlag = parseInt(termModFlags.slice(1, 2));
		if (NtermFlag !== 0 && NtermFlag !== 1) throw "Invalid N-terminal modification flag.";
		if (CtermFlag !== 0 && CtermFlag !== 1) throw "Invalid C-terminal modification flag.";

		var modFlags = new Array(chain.length);
		for (var i = 0; i < chain.length; i++)
		{
			modFlags[i] = 0;
			var c = chain.charAt(i);
			if (c === '1' || c === '2' || c === '3')
			{
				modFlags[i] = c - '0';
				chain = chain.slice(0, i) + chain.slice(i + 1, chain.length); // Remove .charAt(i)
			}
		}

		if (chainNo !== 0)
			renderChainBackground(chainNo, NtermFlag, CtermFlag, residueIndex, residueIndex + chain.length - 1);

		var virtX = (residueIndex % lettersPerLine) * cellSize;
		var virtY = Math.floor(residueIndex / lettersPerLine) * (cellSize + lineSpacing);

		for (var i = 0; i < chain.length; i++)
		{
			var modFlag = modFlags[i] - '0';
			if ((modFlag & 2) != 0) // & SEQ_RNDR_BITFLAG_DFORM
			{
				var savedStrokeStyle = ctx.strokeStyle;
				var savedLineWidth   = ctx.lineWidth;

				ctx.closePath();
				ctx.beginPath();
				ctx.strokeStyle = "#ffffff";
				ctx.lineWidth = Math.max(xToPixels(cellSize) / 6, 3);

				ctx.moveTo(xToScreen(virtX + cellSize / 5), yToScreen(virtY + cellSize * 4 / 5));
				ctx.lineTo(xToScreen(virtX + cellSize * 4 / 5), yToScreen(virtY + cellSize / 5));

				ctx.stroke();
				ctx.closePath();

				ctx.strokeStyle = savedStrokeStyle;
				ctx.lineWidth   = savedLineWidth;
			}

			ctx.fillStyle = "#000000";
			if ((modFlag & 1) != 0) // & SEQ_RNDR_BITFLAG_MODRES
				ctx.fillStyle = "#ff0000";

			drawText( virtX, virtY, chain.slice(i, i + 1) );

			virtX += cellSize;
			if (virtX >= cellSize * lettersPerLine)
			{
				virtX = 0;
				virtY += cellSize + lineSpacing;
			}
		}

		residueIndex += chain.length;
	}
	
	function renderLine(values_to_skip, lineColor)
	{
		var savedStrokeStyle = ctx.strokeStyle;
		var savedLineWidth   = ctx.lineWidth;
		
		var values = currentToken.split(",");

		ctx.closePath();
		ctx.beginPath();
		ctx.strokeStyle = lineColor;
		ctx.lineWidth = Math.max(xToPixels(cellSize) / 4, 2);

		// Start at index 'values_to_skip' (discarding 'from' and 'to' and other non-line-drawing information).
		var valueIdx = values_to_skip;
		var x, y;
		x = xToScreen( parseInt(values[valueIdx]) );
		y = yToScreen( parseInt(values[valueIdx + 1]) );
		ctx.moveTo(x, y);

		valueIdx += 2;
		while (valueIdx < values.length) {
			x = xToScreen( parseInt(values[valueIdx]) );
			y = yToScreen( parseInt(values[valueIdx + 1]) );
			ctx.lineTo(x, y);
			valueIdx += 2;
		}
		ctx.stroke();
		ctx.closePath();
		
		ctx.strokeStyle = savedStrokeStyle;
		ctx.lineWidth   = savedLineWidth;
	}

	// ============ Main function ====================
	
	try {
	
	var render_tokens = render_info.split(':');
	// There must be at least 4 elements: A blank, a header, a bounding box, and an end tag.
	if (render_tokens.length < 4)
		throw 'No rendering info or malformed rendering info.';
	if (render_tokens[1] != 'BCF_SEQ_RNDR_INFO_V1')
		throw 'Rendering info does not start with an "BCF_SEQ_RNDR_INFO_V1" header tag.';
	if (render_tokens[render_tokens.length - 1] != 'ED')
		throw 'Rendering info does not end with an "ED" end tag.';

	var canvas = document.getElementById(canvas_id);
	if (!canvas.getContext)
		return;
	
	var ctx = canvas.getContext('2d');
	var seqFont = font_height + "px 'Courier New'";
	
	ctx.font = seqFont;

	// Read bounding box values.
	var values = render_tokens[2].slice(2).split(',');
	
	var cellSize       = parseInt(values[0]);
	var lineSpacing    = parseInt(values[1]);
	var lettersPerLine = parseInt(values[2]);
	var numRows        = parseInt(values[3]);
	var residueCount   = parseInt(values[4]);

	// Add a little breathing space below characters.
	var fontMetricsHeight = font_height * 1.2;
	var fontMetricsWidth  = ctx.measureText('W').width;
	var charAscent = ctx.measureText('m').width; // Hack - an approximation found at http://stackoverflow.com/questions/1134586/how-can-you-find-the-height-of-text-on-an-html-canvas.

	var xScale  = fontMetricsWidth / cellSize * (1 + lineSpacing / cellSize);
	var yScale  = fontMetricsHeight / cellSize;
	var xOffset = fontMetricsWidth * (MAX_NUM_DIGITS + 2);

	// Bounding box size in pixels.
	var xBBoxSize = xToScreen(cellSize * lettersPerLine);
	var yBBoxSize = yToScreen((cellSize + lineSpacing) * numRows);
	// Add some breathing space.
	xBBoxSize = xBBoxSize + xToPixels(cellSize * 2);
	yBBoxSize = yBBoxSize + yToPixels(lineSpacing * 2);

	if (scale_canvas) {
		canvas.width  = xBBoxSize;
		canvas.height = yBBoxSize;
		// Set font again, it will have been reset by above.
		ctx.font = seqFont;
	}

	//(**!!renderTestGraphics(true, false);	// DEBUG ONLY

	renderNumbers();

	var residueIndex   = 0;
	var disulfideCount = 0;
	var cycleCount     = 0;
	
	// Process all tokens - first 3 have been done, last one has been checked and it is an end-tag.
	for (var tokenIdx = 3; tokenIdx < render_tokens.length - 1; tokenIdx++) {
		var currentToken = render_tokens[tokenIdx];
		var tag = currentToken.slice(0, 2);
		currentToken = currentToken.slice(2);

		if (tag === "CH")
			renderChain();
		else if (tag === "SS") {
			renderLine(2, DISULFIDE_COLORS[disulfideCount % NUM_DISULFIDE_COLORS]);
			disulfideCount++;
		}
		else
		if (tag === "XL") {
			// Check validity of cycle terminal flags, but ignore them.
			if (currentToken.charAt(0) !== '0' && currentToken.charAt(0) !== '1') throw 'Invalid cyclization terminal flag.';
			if (currentToken.charAt(1) !== '0' && currentToken.charAt(1) !== '1') throw 'Invalid cyclization terminal flag.';
			if (currentToken.charAt(2) !== ',') throw 'Invalid cyclization terminal flag end delimiter.';
			renderLine(3, CYCLE_COLORS[cycleCount % NUM_CYCLE_COLORS]);
			cycleCount++;
		}
		else
			throw Error('Unsupported render info tag "' + tag + '".');
	}
			
	} catch (e) {
		ctx.fillStyle = "#ff0000";
		ctx.strokeStyle = "#000000";
		drawText(10, 10, e);
	}
}
